# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import datetime
import logging
from collections import defaultdict
from collections.abc import Callable
from functools import cached_property
from pathlib import Path
from typing import Any
from typing import Literal

import deprecation
import earthkit.data as ekd
from anemoi.utils.checkpoints import load_metadata
from earthkit.data.utils.dates import to_datetime

from anemoi.inference._version import __version__
from anemoi.inference.types import DataRequest
from anemoi.inference.types import Date

from .metadata import Metadata
from .metadata import Variable

LOG = logging.getLogger(__name__)


def _download_huggingfacehub(huggingface_config: Any) -> str:
    """Download model from huggingface.

    Parameters
    ----------
    huggingface_config : dict or str
        Configuration for downloading from huggingface.

    Returns
    -------
    str
        Path to the downloaded model.
    """
    try:
        from huggingface_hub import hf_hub_download
        from huggingface_hub import snapshot_download
    except ImportError as e:
        raise ImportError("Could not import `huggingface_hub`, please run `pip install huggingface_hub`.") from e

    if isinstance(huggingface_config, str):
        huggingface_config = {"repo_id": huggingface_config}

    if "filename" in huggingface_config:
        return str(hf_hub_download(**huggingface_config))

    repo_path = Path(snapshot_download(**huggingface_config))
    ckpt_files = list(repo_path.glob("*.ckpt"))

    if len(ckpt_files) == 1:
        return str(ckpt_files[0])
    else:
        raise ValueError(
            f"None or Multiple ckpt files found in repo, {ckpt_files}.\nCannot pick one to load, please specify `filename`."
        )


class Checkpoint:
    """Represents an inference checkpoint."""

    def __init__(
        self,
        source: str | Metadata | dict[str, Any],
        *,
        patch_metadata: dict[str, Any] | None = None,
    ) -> None:
        """Initialize the Checkpoint.

        Parameters
        ----------
        path : str
            The path to the checkpoint.
        patch_metadata : Optional[Dict[str, Any]], optional
            Metadata to patch the checkpoint with, by default None.
        """
        self._source = source
        self.patch_metadata = patch_metadata

    def __repr__(self) -> str:
        """Represent the Checkpoint as a string.

        Returns
        -------
        str
            String representation of the Checkpoint.
        """
        return f"Checkpoint({self.path})"

    @cached_property
    def path(self) -> str:
        """Get the path to the checkpoint."""
        import json

        try:
            path = json.loads(self._source)
        except Exception:
            path = self._source

        if isinstance(path, (Path, str)):
            return str(path)
        elif isinstance(path, dict):
            if "huggingface" in path:
                return _download_huggingfacehub(path["huggingface"])
            pass
        raise TypeError(f"Cannot parse model path: {path}. It must be a path or dict")

    @cached_property
    def _metadata(self) -> Metadata:
        """Get the metadata."""

        if isinstance(self._source, Metadata):
            return self._source

        try:
            result = Metadata(*load_metadata(self.path, supporting_arrays=True))
        except Exception as e:
            LOG.warning("Version does not support `supporting_arrays` (%s)", e)
            result = Metadata(load_metadata(self.path))

        if self.patch_metadata:
            LOG.warning("Patching metadata with %r", self.patch_metadata)
            result.patch(self.patch_metadata)

        return result

    ###########################################################################
    # Forwards used by the runner
    # We do not want to expose the metadata object directly
    # We do not use `getattr` to avoid exposing all methods and make debugging
    # easier
    ###########################################################################

    @property
    def timestep(self) -> Any:
        """Get the timestep."""
        return self._metadata.timestep

    @property
    def input_explicit_times(self) -> Any:
        """Get the input explicit times from metadata."""
        return self._metadata.input_explicit_times

    @property
    def target_explicit_times(self) -> Any:
        """Get the target explicit times."""
        return self._metadata.target_explicit_times

    @property
    def data_frequency(self) -> Any:
        """Get the data frequency."""
        return self._metadata._config_data.frequency

    @property
    def precision(self) -> Any:
        """Get the precision."""
        return self._metadata.precision

    @property
    def number_of_grid_points(self) -> Any:
        """Get the number of grid points."""
        return self._metadata.number_of_grid_points

    @property
    def number_of_input_features(self) -> Any:
        """Get the number of input features."""
        return self._metadata.number_of_input_features

    @property
    def variable_to_input_tensor_index(self) -> Any:
        """Get the variable to input tensor index."""
        return self._metadata.variable_to_input_tensor_index

    @property
    def model_computed_variables(self) -> Any:
        """Get the model computed variables."""
        return self._metadata.model_computed_variables

    @property
    def typed_variables(self) -> dict[str, Variable]:
        """Get the typed variables."""
        return self._metadata.typed_variables

    @property
    @deprecation.deprecated(
        deprecated_in="0.6.4",
        removed_in="0.8.0",
        current_version=__version__,
        details="Use `select_variables` instead.",
    )
    def diagnostic_variables(self) -> Any:
        """Get the diagnostic variables."""
        return self._metadata.diagnostic_variables

    @property
    @deprecation.deprecated(
        deprecated_in="0.6.4",
        removed_in="0.8.0",
        current_version=__version__,
        details="Use `select_variables` instead.",
    )
    def prognostic_variables(self) -> Any:
        """Get the prognostic variables."""
        return self._metadata.prognostic_variables

    @property
    def prognostic_output_mask(self) -> Any:
        """Get the prognostic output mask."""
        return self._metadata.prognostic_output_mask

    @property
    def prognostic_input_mask(self) -> Any:
        """Get the prognostic input mask."""
        return self._metadata.prognostic_input_mask

    @property
    def output_tensor_index_to_variable(self) -> Any:
        """Get the output tensor index to variable."""
        return self._metadata.output_tensor_index_to_variable

    @property
    def accumulations(self) -> Any:
        """Get the accumulations."""
        return self._metadata.accumulations

    @property
    def latitudes(self) -> Any:
        """Get the latitudes."""
        return self._metadata.latitudes

    @property
    def longitudes(self) -> Any:
        """Get the longitudes."""
        return self._metadata.longitudes

    @property
    def grid_points_mask(self) -> Any:
        """Get the grid points mask."""
        return self._metadata.grid_points_mask

    @cached_property
    def sources(self) -> list["SourceCheckpoint"]:
        """Get the sources."""
        return [SourceCheckpoint(self, _) for _ in self._metadata.sources(self.path)]

    def default_namer(self, *args: Any, **kwargs: Any) -> Callable[[ekd.Field, Any], str]:
        """Return a callable that can be used to name fields.

        Parameters
        ----------
        *args : Any
            Additional arguments.

        **kwargs : Any
            Additional keyword arguments.

        Returns
        -------
        Callable
            The namer that was used to create the training dataset.
        """
        return self._metadata.default_namer(*args, **kwargs)

    def report_error(self) -> None:
        """Report an error."""
        self._metadata.report_error()

    def validate_environment(
        self,
        *,
        all_packages: bool = False,
        on_difference: Literal["warn", "error", "ignore", "return"] = "warn",
        exempt_packages: list[str] | None = None,
    ) -> bool | str:
        """Validate the environment.

        Parameters
        ----------
        all_packages : bool, optional
            Check all packages in the environment (True) or just anemoi's (False), by default False.
        on_difference : Literal['warn', 'error', 'ignore', 'return'], optional
            What to do on difference, by default "warn"
        exempt_packages : list[str], optional
            List of packages to exempt from the check, by default EXEMPT_PACKAGES

        Returns
        -------
        Union[bool, str]
            boolean if `on_difference` is not 'return', otherwise formatted text of the differences
            True if environment is valid, False otherwise
        """
        return self._metadata.validate_environment(
            all_packages=all_packages, on_difference=on_difference, exempt_packages=exempt_packages
        )

    def open_dataset(
        self,
        *,
        use_original_paths: bool | None = None,
        from_dataloader: Any | None = None,
    ) -> Any:
        """Open the dataset.

        Parameters
        ----------
        use_original_paths : bool, optional
            Whether to use the original paths, by default None.
        from_dataloader : Any, optional
            The dataloader to use, by default None.

        Returns
        -------
        Any
            The opened dataset.
        """
        return self._metadata.open_dataset(use_original_paths=use_original_paths, from_dataloader=from_dataloader)

    def open_dataset_args_kwargs(
        self, *, use_original_paths: bool, from_dataloader: Any | None = None
    ) -> tuple[Any, Any]:
        """Get arguments and keyword arguments for opening the dataset.

        Parameters
        ----------
        use_original_paths : bool
            Whether to use original paths.
        from_dataloader : Optional[Any], optional
            Data loader, by default None.

        Returns
        -------
        Tuple[Any, Any]
            Arguments and keyword arguments for opening the dataset.
        """
        return self._metadata.open_dataset_args_kwargs(
            use_original_paths=use_original_paths,
            from_dataloader=from_dataloader,
        )

    def name_fields(self, fields: Any, namer: Callable[..., str] | None = None) -> Any:
        """Name fields.

        Parameters
        ----------
        fields : Any
            The fields to name.
        namer : Optional[Callable[...,str]], optional
            The namer, by default None.

        Returns
        -------
        Any
            The named fields.
        """
        return self._metadata.name_fields(fields, namer=namer)

    def sort_by_name(
        self, fields: ekd.FieldList, *args: Any, namer: Callable[..., str] | None = None, **kwargs: Any
    ) -> ekd.FieldList:
        """Sort fields by name.

        Parameters
        ----------
        fields : ekd.FieldList
            The fields to sort.
        *args : Any
            Additional arguments.
        namer : Optional[Callable[...,str]], optional
            The namer, by default None.
        **kwargs : Any
            Additional keyword arguments.

        Returns
        -------
        ekd.FieldList
            The sorted fields.
        """
        return self._metadata.sort_by_name(fields, *args, namer=namer, **kwargs)

    def print_indices(self, print=LOG.info) -> None:
        """Print the indices."""
        return self._metadata.print_indices(print=print)

    ###########################################################################
    # Variable categories
    ###########################################################################
    def print_variable_categories(self, print=LOG.info) -> None:
        """Print the variable categories."""
        return self._metadata.print_variable_categories(print=print)

    def variable_categories(self) -> Any:
        """Get the variable categories.

        Returns
        -------
        Any
            The variable categories.
        """
        return self._metadata.variable_categories()

    def select_variables(self, *, include: list[str] | None = None, exclude: list[str] | None = None) -> list[str]:
        """Get variables from input.

        Parameters
        ----------
        include : Optional[List[str]]
            Categories to include.

        exclude : Optional[List[str]]
            Categories to exclude.

        Returns
        -------
        List[str]
            The selected variables.

        """
        return self._metadata.select_variables(include=include, exclude=exclude)

    def select_variables_and_masks(
        self, *, include: list[str] | None = None, exclude: list[str] | None = None
    ) -> list[str]:
        """Get variables from input.

        Parameters
        ----------
        include : Optional[List[str]]
            Categories to include.

        exclude : Optional[List[str]]
            Categories to exclude.

        Returns
        -------
        List[str]
            The selected variables.
        """
        return self._metadata.select_variables_and_masks(include=include, exclude=exclude)

    ###########################################################################
    def load_supporting_array(self, name: str) -> Any:
        """Load a supporting array.

        Parameters
        ----------
        name : str
            The name of the supporting array.

        Returns
        -------
        Any
            The supporting array.
        """
        return self._metadata.load_supporting_array(name)

    @property
    def supporting_arrays(self) -> Any:
        """Get the supporting arrays."""
        return self._metadata.supporting_arrays

    ###########################################################################

    @cached_property
    def lagged(self) -> list[datetime.timedelta]:
        """Return the list of steps for the `multi_step_input` fields."""
        result = list(range(0, self._metadata.multi_step_input))

        result = [-s * self._metadata.timestep for s in result]
        return sorted(result)

    @property
    def multi_step_input(self) -> Any:
        """Get the multi-step input."""
        return self._metadata.multi_step_input

    ###########################################################################
    # Data retrieval
    ###########################################################################
    @property
    def grid(self) -> Any:
        """Get the grid."""
        return self._metadata.grid

    @property
    def area(self) -> Any:
        """Get the area."""
        return self._metadata.area

    def mars_by_levtype(self, levtype: str) -> Any:
        """Get MARS requests by level type.

        Parameters
        ----------
        levtype : str
            The level type.

        Returns
        -------
        Any
            The MARS requests by level type.
        """
        return self._metadata.mars_by_levtype(levtype)

    def mars_requests(
        self,
        *,
        variables: list[str],
        dates: list[Date],
        use_grib_paramid: bool = False,
        always_split_time: bool = False,
        patch_request: Callable[[DataRequest], DataRequest] | None = None,
        dont_fail_for_missing_paramid: bool = False,
        **kwargs: Any,
    ) -> list[DataRequest]:
        """Generate MARS requests for the given variables and dates.

        Parameters
        ----------
        variables : List[str]
            The list of variables.
        dates : List[Any]
            The list of dates.
        use_grib_paramid : bool, optional
            Whether to use GRIB paramid, by default False.
        always_split_time : bool, optional
            Whether to always split time, by default False.
        patch_request : Optional[Callable], optional
            A callable to patch the request, by default None.
        dont_fail_for_missing_paramid : bool, optional
            Whether to not fail for missing param ids, by default False.
        **kwargs : Any
            Additional keyword arguments.

        Returns
        -------
        List[DataRequest]
            The list of MARS requests.
        """
        from anemoi.utils.grib import shortname_to_paramid
        from earthkit.data.utils.availability import Availability

        assert variables, "No variables provided"

        if not isinstance(dates, (list, tuple)):
            dates = [dates]

        dates = [to_datetime(d) for d in dates]

        assert dates, "No dates provided"

        result: list[DataRequest] = []

        DEFAULT_KEYS = ("class", "expver", "type", "stream", "levtype")
        DEFAULT_KEYS_AND_TIME = ("class", "expver", "type", "stream", "levtype", "time")

        # ECMWF operational data has stream oper for 00 and 12 UTC and scda for 06 and 18 UTC
        # The split oper/scda is a bit special
        KEYS = {("oper", "fc"): DEFAULT_KEYS_AND_TIME, ("scda", "fc"): DEFAULT_KEYS_AND_TIME}

        requests = defaultdict(list)
        for r in self._metadata.mars_requests(variables=variables):
            for date in dates:

                r = r.copy()

                base = date
                step = str(r.get("step", 0)).split("-")[-1]
                step = int(step)
                base = base - datetime.timedelta(hours=step)

                r["date"] = base.strftime("%Y-%m-%d")
                r["time"] = base.strftime("%H%M")

                r.update(kwargs)  # We do it here so that the Availability can use that information

                if always_split_time:
                    keys = DEFAULT_KEYS_AND_TIME
                else:
                    keys = KEYS.get((r.get("stream"), r.get("type")), DEFAULT_KEYS)
                key = tuple(r.get(k) for k in keys)

                # Special case because of oper/scda

                requests[key].append(r)

        result = []
        for reqs in requests.values():

            compressed = Availability(reqs)
            for r in compressed.iterate():

                if not r:
                    continue

                r = r.copy()

                # Convert all to lists
                for k, v in r.items():
                    if not isinstance(v, (list, tuple, set)):
                        v = [v]
                    r[k] = sorted(set(v))

                # Patch BEFORE the shortname to paramid

                if patch_request:
                    r = patch_request(r)

                # Convert all to lists (again)
                for k, v in r.items():
                    if not isinstance(v, (list, tuple, set)):
                        v = [v]
                    r[k] = sorted(set(v))

                if use_grib_paramid and "param" in r:

                    def shortname_to_paramid_no_fail(x: str) -> str:
                        try:
                            return shortname_to_paramid(x)
                        except KeyError:
                            LOG.warning("Could not convert shortname '%s' to paramid", x)
                            return x

                    if dont_fail_for_missing_paramid:
                        _ = shortname_to_paramid_no_fail
                    else:
                        _ = shortname_to_paramid

                    r["param"] = [_(p) for p in r["param"]]

                # Simplify the request

                for k in list(r.keys()):
                    v = r[k]
                    if len(v) == 1:
                        v = v[0]

                    # Remove empty values for when tree is not fully defined
                    if v == "-":
                        r.pop(k)
                        continue
                    r[k] = v
                result.append(r)

        return result

    ###########################################################################
    # supporting arrays
    ###########################################################################

    @cached_property
    def _supporting_arrays(self) -> Any:
        """Get the supporting arrays."""
        return self._metadata._supporting_arrays

    @property
    def name(self) -> Any:
        """Get the name."""
        return self._metadata.name

    def has_supporting_array(self, name: str) -> bool:
        """Check if the checkpoint has a supporting array with the given name.

        Parameters
        ----------
        name : str
            The name of the supporting array.

        Returns
        -------
        bool
            True if the supporting array exists, False otherwise.
        """
        return self._metadata.has_supporting_array(name)

    ###########################################################################
    # Misc
    ###########################################################################

    def provenance_training(self) -> Any:
        """Get the provenance of the training.

        Returns
        -------
        Any
            The provenance of the training.
        """
        return self._metadata.provenance_training()


class SourceCheckpoint(Checkpoint):
    """A checkpoint that represents a source."""

    def __init__(self, owner: Checkpoint, metadata: Any) -> None:
        """Initialize the SourceCheckpoint.

        Parameters
        ----------
        owner : Checkpoint
            The owner checkpoint.
        metadata : Any
            The metadata for the source checkpoint.
        """
        super().__init__(owner.path)
        self._owner = owner
        self._metadata = metadata

    def __repr__(self) -> str:
        """Represent the SourceCheckpoint as a string.

        Returns
        -------
        str
            String representation of the SourceCheckpoint.
        """
        return f"Source({self.name}@{self.path})"

    @property
    def operational_config(self) -> dict[str, Any]:
        LOG.warning("The `operational_config` property is deprecated.")
        return False



# (C) Copyright 2025- Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from pathlib import Path

import numpy as np

from anemoi.inference.context import Context
from anemoi.inference.decorators import main_argument
from anemoi.inference.types import BoolArray
from anemoi.inference.types import State

from ..processor import Processor
from . import post_processor_registry

LOG = logging.getLogger(__name__)


class ExtractBase(Processor):
    """Base class for processors that extract data from the state."""

    # this needs to be set in subclasses
    indexer: BoolArray | slice

    def process(self, state: State) -> State:
        """Process the state to extract a subset of points based on the indexer.

        Parameters
        ----------
        state : State
            The state containing fields to be extracted.

        Returns
        -------
        State
            The updated state with extracted fields.
        """
        state = state.copy()
        state["fields"] = state["fields"].copy()

        state["latitudes"] = state["latitudes"][self.indexer]
        state["longitudes"] = state["longitudes"][self.indexer]
        for field in state["fields"]:
            state["fields"][field] = state["fields"][field][self.indexer]

        return state


@post_processor_registry.register("extract_mask")
@main_argument("mask")
class ExtractMask(ExtractBase):
    """Extract a subset of points from the state based on a boolean mask.

    Parameters
    ----------
    context : Any
        The context in which the processor is running.
    mask : str
        Either a path to a `.npy` file containing the boolean mask or
        the name of a supporting array found in the checkpoint.
    """

    def __init__(self, context: Context, mask: str) -> None:
        super().__init__(context)

        self._maskname = mask

        if Path(mask).is_file():
            mask = np.load(mask)
        else:
            mask = context.checkpoint.load_supporting_array(mask)

        if not isinstance(mask, np.ndarray) or mask.dtype != bool:
            raise ValueError(
                "Expected the mask to be a boolean numpy array. " f"Got {type(mask)} with dtype {mask.dtype}."
            )

        self.indexer = mask
        self.npoints = np.sum(mask)

    def __repr__(self) -> str:
        """Return a string representation of the ExtractMask object.

        Returns
        -------
        str
            String representation of the object.
        """
        return f"ExtractMask({self._maskname}, points={self.npoints}/{self.indexer.size})"


@post_processor_registry.register("extract_slice")
class ExtractSlice(ExtractBase):
    """Extract a subset of points from the state based on a slice.

    Parameters
    ----------
    context : Context
        The context in which the processor is running.
    slice_args : int
        Arguments to create a slice object. This can be a single integer or
        a tuple of integers representing the start, stop, and step of the slice.
    """

    def __init__(self, context: Context, *slice_args: int) -> None:
        super().__init__(context)
        self.indexer = slice(*slice_args)

    def __repr__(self) -> str:
        """Return a string representation of the ExtractSlice object.

        Returns
        -------
        str
            String representation of the object.
        """
        return f"ExtractSlice({self.indexer})"


@post_processor_registry.register("extract_from_state")
@main_argument("cutout_source")
class ExtractState(ExtractBase):
    """Extract a subset of points based on a mask included in the state.

    Must be used with the Cutout input.
    """

    def __init__(self, context: Context, cutout_source: str) -> None:
        """Initialize the ExtractState processor.

        Must be used with the Cutout input.

        Parameters
        ----------
        context : Context
            The context in which the processor is running.
        cutout_source : str
            The source to use for the cutout mask for extraction.
        """
        super().__init__(context)
        self._source = cutout_source

    def process(self, state: State) -> State:
        """Process the state to extract a subset of points based on the mask.

        Parameters
        ----------
        state : State
            The state containing fields to be extracted.

        Returns
        -------
        State
            The updated state with extracted fields.
        """
        if "_mask" not in state:
            raise ValueError("Private attribute '_mask' not found in the state. Are you using the Cutout Input?")

        if self._source not in state["_mask"]:
            raise ValueError(f"Mask '{self._source}' not found in the state.")

        mask = state["_mask"][self._source]
        if not isinstance(mask, np.ndarray) or mask.dtype != bool:
            raise ValueError(f"Expected the mask '{self._source}' to be a boolean numpy array.")

        self.indexer = mask
        extraced_state = super().process(state)

        for key in (k for k in state.keys() if k.startswith("_")):
            if isinstance(state[key], dict) and self._source in state[key]:
                extraced_state[key] = state[key][self._source]

        return extraced_state

    def __repr__(self) -> str:
        """Return a string representation of the ExtractSlice object.

        Returns
        -------
        str
            String representation of the object.
        """
        return f"ExtractState(source={self._source!r})"



# (C) Copyright 2025- Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import re

_TEMPLATE_EXPRESSION_PATTERN = re.compile(r"\{(.*?)\}")


def render_template(template: str, handle: dict) -> str:
    """Render a template string with the given keyword arguments.

    Given a template string such as '{dateTime}_{step:03}.grib' and
    the GRIB handle, this function will replace the expressions in the
    template with the corresponding values from the handle, formatted
    according to the optional format specifier.

    For example, the template '{dateTime}_{step:03}.grib' with a handle
    containing 'dateTime' as '202501011200' and 'step' as 6 will
    produce '202501011200_006.grib'.

    Parameters
    ----------
    template : str
        The template string to render.
    handle : dict
        The dictionary to use for rendering the template.

    Returns
    -------
    str
        The rendered template string.
    """
    expressions = _TEMPLATE_EXPRESSION_PATTERN.findall(str(template))
    expr_format = [el.split(":") if ":" in el else [el, ""] for el in expressions]
    keys = {k[0]: format(handle.get(k[0]), k[1]) for k in expr_format}
    path = str(template).format(**keys)
    return path



# (C) Copyright 2025 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from typing import Any

import earthkit.data as ekd
import yaml
from anemoi.utils.registry import Registry

from anemoi.inference.config import Configuration

LOG = logging.getLogger(__name__)


template_provider_registry = Registry(__name__)


def create_template_provider(owner: Any, config: Configuration) -> "TemplateProvider":
    """Create a template provider from the given configuration.

    Parameters
    ----------
    owner : Any
        The owner of the template provider.
    config : Configuration
        The configuration for the template provider.

    Returns
    -------
    TemplateProvider
        The created template provider.
    """
    return template_provider_registry.from_config(config, owner)


class TemplateProvider:
    """Base class for template providers."""

    def __init__(self, manager: Any) -> None:
        """Initialize the template provider.

        Parameters
        ----------
        manager : Any
            The manager for the template provider.
        """
        self.manager = manager

    def template(self, variable: str, lookup: dict[str, Any]) -> ekd.Field:
        """Get the template for the given variable and lookup.

        Parameters
        ----------
        variable : str
            The variable to get the template for.
        lookup : Dict[str, Any]
            The lookup dictionary.

        Returns
        -------
        ekd.Field
            The template field.
        """
        raise NotImplementedError()


class IndexTemplateProvider(TemplateProvider):
    """Template provider based on an index file."""

    def __init__(self, manager: Any, index_path: str) -> None:
        """Initialize the index template provider.

        Parameters
        ----------
        manager : Any
            The manager for the template provider.
        index_path : str
            The path to the index file.
        """
        super().__init__(manager)
        self.index_path = index_path

        with open(index_path) as f:
            self.templates = yaml.safe_load(f)

        if not isinstance(self.templates, list):
            raise ValueError("Invalid templates.yaml, must be a list")

        # TODO: use pydantic
        for template in self.templates:
            if not isinstance(template, list):
                raise ValueError(f"Invalid template in templates.yaml, must be a list: {template}")
            if len(template) != 2:
                raise ValueError(f"Invalid template in templates.yaml, must have exactly 2 elements: {template}")

            match, grib = template
            if not isinstance(match, dict):
                raise ValueError(f"Invalid match in templates.yaml, must be a dict: {match}")

            if not isinstance(grib, str):
                raise ValueError(f"Invalid grib in templates.yaml, must be a string: {grib}")

    def template(self, variable: str, lookup: dict[str, Any]) -> ekd.Field | None:
        """Get the template for the given variable and lookup.

        Parameters
        ----------
        variable : str
            The variable to get the template for.
        lookup : Dict[str, Any]
            The lookup dictionary.

        Returns
        -------
        Optional[ekd.Field]
            The template field if found, otherwise None.
        """

        def _as_list(value: Any) -> list[Any]:
            if not isinstance(value, list):
                return [value]
            return value

        for template in self.templates:
            match, grib = template
            if LOG.isEnabledFor(logging.DEBUG):
                LOG.debug("%s", [(lookup.get(k), _as_list(v)) for k, v in match.items()])

            if all(lookup.get(k) in _as_list(v) for k, v in match.items()):
                return self.load_template(grib, lookup)

        return None

    def load_template(self, grib: str, lookup: dict[str, Any]) -> ekd.Field | None:
        """Load the template for the given GRIB and lookup.

        Parameters
        ----------
        grib : str
            The GRIB string.
        lookup : Dict[str, Any]
            The lookup dictionary.

        Returns
        -------
        Optional[ekd.Field]
            The loaded template field if found, otherwise None.
        """
        raise NotImplementedError()



# (C) Copyright 2025 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import torch


def get_available_device() -> "torch.device":
    """Get the available device for PyTorch.

    Returns
    -------
    torch.device
        The available device, either 'cuda', 'mps', or 'cpu'.
    """
    import torch

    if torch.cuda.is_available():
        return torch.device("cuda")
    elif torch.backends.mps.is_available():
        return torch.device("mps")
    return torch.device("cpu")



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
import os
from functools import cached_property
from typing import Any

import earthkit.data as ekd

from anemoi.inference.context import Context
from anemoi.inference.types import Date
from anemoi.inference.types import State

from ..decorators import main_argument
from . import input_registry
from .grib import GribInput

LOG = logging.getLogger(__name__)


@input_registry.register("grib")
@main_argument("path")
class GribFileInput(GribInput):
    """Handles grib files."""

    trace_name = "grib file"

    def __init__(
        self,
        context: Context,
        *,
        path: str,
        **kwargs: Any,
    ) -> None:
        """Initialize the GribFileInput.

        Parameters
        ----------
        context : Any
            The context in which the input is used.
        path : str
            The path to the GRIB file.
        pre_processors : Optional[List[ProcessorConfig]], default None
            Pre-processors to apply to the input
        namer : Optional[Any]
            Optional namer for the input.
        **kwargs : Any
            Additional keyword arguments.
        """
        super().__init__(context, **kwargs)
        self.path = path

    def create_input_state(self, *, date: Date | None) -> State:
        """Create the input state for the given date.

        Parameters
        ----------
        date : Optional[Date]
            The date for which to create the input state.

        Returns
        -------
        State
            The created input state.
        """

        return self._create_input_state(self._fieldlist, date=date)

    def load_forcings_state(self, *, dates: list[Date], current_state: State) -> State:
        """Load the forcings state for the given variables and dates.

        Parameters
        ----------
        dates : List[Date]
            List of dates for which to load the forcings.
        current_state : State
            The current state of the input.

        Returns
        -------
        State
            The loaded forcings state.
        """

        return self._load_forcings_state(
            self._fieldlist,
            dates=dates,
            current_state=current_state,
        )

    @cached_property
    def _fieldlist(self) -> ekd.FieldList:
        """Get the input fieldlist from the GRIB file."""
        if os.path.getsize(self.path) == 0:
            LOG.warning("GRIB file %r is empty", self.path)
            return ekd.from_source("empty")
        return ekd.from_source("file", self.path)



# (C) Copyright 2025 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

"""Dummy input used for testing.

It will generate fields with constant values for each variable and date.
These values are then tested in the mock model.
"""

import logging
from typing import Any

from anemoi.inference.context import Context
from anemoi.inference.types import Date
from anemoi.inference.types import State

from ..input import Input
from . import input_registry

LOG = logging.getLogger(__name__)
SKIP_KEYS = ["date", "time", "step"]


@input_registry.register("empty")
class EmptyInput(Input):
    """An Input that is always empty."""

    trace_name = "empty"

    def __init__(self, context: Context, **kwargs: Any) -> None:
        """Initialise the EmptyInput.

        Parameters
        ----------
        context : Context
            The context object for the input.
        **kwargs : object
            Additional keyword arguments.
        """
        super().__init__(context, **kwargs)
        assert self.variables in (None, []), "EmptyInput should not have variables"

    def create_input_state(self, *, date: Date | None) -> State:
        """Create an empty input state.

        Parameters
        ----------
        date : Date or None
            The date for the input state.

        Returns
        -------
        State
            The created empty input state.
        """
        return dict(fields=dict(), _input=self)

    def load_forcings_state(self, *, dates: list[Date], current_state: State) -> State:
        """Load an empty forcings state.

        Parameters
        ----------
        dates : list of Date
            The list of dates for the forcings state.
        current_state : State
            The current state (unused).

        Returns
        -------
        State
            The loaded empty forcings state.
        """
        return dict(date=dates[-1], fields=dict(), _input=self)



# (C) Copyright 2024-2025 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
import warnings
from io import IOBase
from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any
from typing import Hashable

import earthkit.data as ekd
from anemoi.utils.dates import as_timedelta

from anemoi.inference.utils.templating import render_template

if TYPE_CHECKING:
    from anemoi.transform.variables import Variable

LOG = logging.getLogger(__name__)


GRIB1_ONLY: list[str] = []

GRIB2_ONLY: list[str] = ["typeOfGeneratingProcess"]


ORDERING = (
    "edition",
    "typeOfLevel",
    "stepType",
    "productDefinitionTemplateNumber",
    "eps",
    "number",
)

ORDERING = {k: i for i, k in enumerate(ORDERING)}


def _ordering(item: tuple) -> int:
    """Get the ordering index for a given item.

    Parameters
    ----------
    item : tuple
        The item to get the ordering index for.

    Returns
    -------
    int
        The ordering index.
    """
    return ORDERING.get(item[0], 999)


def _param(param: Any) -> str:
    """Determine the parameter type based on its value.

    Parameters
    ----------
    param : str
        The parameter value.

    Returns
    -------
    str
        The parameter type.
    """
    try:
        int(param)
        return "paramId"
    except ValueError:
        try:
            float(param)
            return "param"
        except ValueError:
            return "shortName"


def _step_in_hours(step: Any) -> int:
    """Convert a step to hours.

    Parameters
    ----------
    step : Any
        The step to convert.

    Returns
    -------
    int
        The step in hours.
    """
    step = step.total_seconds() / 3600
    assert int(step) == step
    return int(step)


STEP_TYPE = {
    "accumulation": "accum",
    "average": "avg",
    "maximum": "max",
    "minimum": "min",
    "instantaneous": None,
}


def encode_time_processing(
    *,
    result: dict[str, Any],
    template: ekd.Field,
    variable: "Variable",
    step: Any,
    previous_step: Any | None,
    start_steps: dict[Any, Any],
    edition: int,
    ensemble: bool,
) -> None:
    """Encode time processing information into the result dictionary.

    Parameters
    ----------
    result : Dict[str, Any]
        The result dictionary to update.
    template : ekd.Field
        The template field.
    variable : Any
        The variable containing time processing information.
    step : Any
        The current step.
    previous_step : Optional[Any]
        The previous step.
    start_steps : Dict[Any, Any]
        The start steps dictionary.
    edition : int
        The GRIB edition.
    ensemble : bool
        Whether the data is part of an ensemble.
    """
    assert edition in (1, 2)

    if variable.time_processing is None:
        result["step"] = _step_in_hours(step)
        result["stepType"] = "instant"
        return

    if previous_step is None:
        if not variable.is_accumulation:
            LOG.warning(f"No previous step available for time processing `{variable.time_processing}` for `{variable}`")
        previous_step = step

    if period := getattr(variable, "period", None):
        start = step - period
        if start < as_timedelta(0):
            raise ValueError(
                f"Writing {variable.name} at step {_step_in_hours(step)} with period {_step_in_hours(period)} for would result in a negative start step {_step_in_hours(start)}. "
                "Try `write_initial_state: False`"
            )
    else:
        # backwards compatibility with old transform or if period is missing from the metadata
        start = previous_step
        warnings.warn(
            f"{variable.name} {variable.time_processing} does not have a period set, using previous_step as start={_step_in_hours(start)}."
        )

    # give post-processors a chance to modify the start step
    start = _step_in_hours(start_steps.get(variable.name, start))
    end = _step_in_hours(step)

    result["startStep"] = start
    result["endStep"] = end
    result["stepType"] = STEP_TYPE[variable.time_processing]

    if edition == 1:
        return

    if ensemble:
        result["productDefinitionTemplateNumber"] = 11
    else:
        result["productDefinitionTemplateNumber"] = 8


LEVTYPES = {
    "pl": "isobaricInhPa",
    "ml": "hybrid",
    "pt": "theta",
    "pv": "potentialVorticity",
}


def grib_keys(
    *,
    values: Any,
    template: Any,
    variable: Any,
    ensemble: bool,
    param: int | float | str | None,
    date: int,
    time: int,
    step: Any,
    previous_step: Any | None,
    start_steps: dict[Any, Any],
    keys: dict[str, Any],
    grib1_keys: dict[int | float | str, dict[str, Any]] = {},
    grib2_keys: dict[int | float | str, dict[str, Any]] = {},
) -> dict[str, Any]:
    """Generate GRIB keys for encoding.

    Parameters
    ----------
    values : Any
        The values to encode.
    template : Any
        The template to use.
    variable : Any
        The variable containing GRIB keys.
    ensemble : bool
        Whether the data is part of an ensemble.
    param : Optional[Union[int, float, str]]
        The parameter value.
    date : int
        The date value.
    time : int
        The time value.
    step : Any
        The current step.
    previous_step : Optional[Any]
        The previous step.
    start_steps : Dict[Any, Any]
        The start steps dictionary.
    keys : Dict[str, Any]
        The initial keys dictionary.
    grib1_keys : Dict[Union[int, float, str], Dict[str, Any]], optional
        Additional GRIB1 keys.
    grib2_keys : Dict[Union[int, float, str], Dict[str, Any]], optional
        Additional GRIB2 keys.

    Returns
    -------
    Dict[str, Any]
        The generated GRIB keys.
    """
    result = keys.copy()

    edition = keys.get("edition")
    if edition is None and template is not None:
        edition = template.metadata("edition")

    if edition is None:
        edition = 1

    if edition == 1:
        for k in GRIB2_ONLY:
            result.pop(k, None)

    if edition == 2:
        for k in GRIB1_ONLY:
            result.pop(k, None)

    result["edition"] = edition

    if param is not None:
        result.setdefault(_param(param), param)

        if edition == 1:
            result.update(grib1_keys.get(param, {}))

        if edition == 2:
            result.update(grib2_keys.get(param, {}))

    result.setdefault("type", "fc")

    if result.get("type") in ("an", "fc"):
        # For organisations that do not use type
        result.setdefault("dataType", result.pop("type"))

    result["date"] = date
    result["time"] = time

    encode_time_processing(
        result=result,
        template=template,
        variable=variable,
        step=step,
        previous_step=previous_step,
        start_steps=start_steps,
        edition=edition,
        ensemble=ensemble,
    )

    for k, v in variable.grib_keys.items():
        if k not in (
            "domain",
            "type",
            "expver",
            "class",
            "param",
            "number",
            "step",
            "date",
            "hdate",
            "time",
            "timespan",
            "valid_datetime",
            "variable",
        ):
            if k == "stream":
                if v in ("oper", "wave"):
                    result.setdefault(k, v)
                continue

            if k == "levtype":
                v = LEVTYPES.get(v)
                if v is None:
                    continue
                k = "typeOfLevel"
            result.setdefault(k, v)

    result = {k: v for k, v in sorted(result.items(), key=_ordering) if v is not None}

    return result


def check_encoding(handle: Any, keys: dict[str, Any], first: bool = True) -> None:
    """Check if the GRIB encoding matches the expected keys.

    Parameters
    ----------
    handle : Any
        The GRIB handle.
    keys : Dict[str, Any]
        The expected keys.
    first : bool, optional
        Whether this is the first check.

    Raises
    ------
    ValueError
        If the GRIB field could not be encoded.
    """

    def same(w: Any, v: Any, k: str) -> bool:
        if type(v) is type(w):
            # Keep `mypy` happy
            equal: bool = v == w
            return equal
        return str(w) == str(v)

    mismatches = {}
    for k, v in keys.items():
        if k == "param":
            try:
                int(v)
                k = "paramId"
            except ValueError:
                try:
                    float(v)
                    k = "param"
                except ValueError:
                    k = "shortName"

        if k == "date":
            v = int(str(v).replace("-", ""))

        if k == "time":
            v = int(v)
            if v < 100:
                v *= 100

        if isinstance(v, int):
            w = handle.get_long(k)
        else:
            w = handle.get(k)

        if not same(w, v, k):
            mismatches[k] = f'Expected "{v}" but got "{w}"'

    if mismatches:

        if first:
            import eccodes
            from earthkit.data.readers.grib.codes import GribCodesHandle

            handle = GribCodesHandle(eccodes.codes_clone(handle._handle), None, None)
            return check_encoding(handle, keys, first=False)

        raise ValueError(f"GRIB field could not be encoded. Mismatches={mismatches}")


def encode_message(
    *,
    values: Any | None,
    template: Any,
    metadata: dict[str, Any],
    check_nans: bool = False,
    missing_value: int | float = 9999,
) -> Any:
    """Encode a GRIB message.

    Parameters
    ----------
    values : Optional[Any]
        The values to encode.
    template : Any
        The template to use.
    metadata : Dict[str, Any]
        The metadata for the GRIB message.
    check_nans : bool, optional
        Whether to check for NaNs in the values.
    missing_value : Union[int, float], optional
        The value to use for missing data.

    Returns
    -------
    Any
        The encoded GRIB handle.
    """
    metadata = metadata.copy()  # avoid modifying the original metadata
    handle = template.handle.clone()

    if check_nans and values is not None:
        import numpy as np

        if np.isnan(values).any():
            # missing_value = np.finfo(values.dtype).max
            missing_value = missing_value
            values = np.nan_to_num(values, nan=missing_value)
            metadata["missingValue"] = missing_value
            metadata["bitmapPresent"] = 1

    if int(metadata.get("deleteLocalDefinition", 0)):
        for k in ("class", "type", "stream", "expver", "setLocalDefinition"):
            metadata.pop(k, None)

    metadata.setdefault("generatingProcessIdentifier", 255)

    LOG.debug("GribOutput.metadata %s", metadata)

    single = {}
    multiple = {}
    for k, v in metadata.items():
        if isinstance(v, (int, float, str, bool)):
            single[k] = v
        else:
            multiple[k] = v

    try:
        # Try to set all metadata at once
        # This is needed when we set multiple keys that are interdependent
        handle.set_multiple(single)
    except Exception as e:
        LOG.error("Failed to set metadata at once: %s", e)
        # Try again, but one by one
        for k, v in single.items():
            handle.set(k, v)

    for k, v in multiple.items():
        handle.set(k, v)

    if values is not None:
        handle.set_values(values)

    return handle


class GribWriter:
    """Write GRIB messages to one or more files."""

    def __init__(self, out: Path | IOBase, split_output: bool = True) -> None:
        """Initialize the GribWriter.

        Parameters
        ----------
        out : Union[Path, IOBase]
            Path or file-like object to write the grib data to.
            If a string, it should be a file path.
            If a file-like object, it should be opened in binary write mode.
        split_output : bool, optional
            Whether to split the output into multiple files.
        """

        if isinstance(out, IOBase) and split_output:
            raise ValueError("Cannot split output when `out` is a file-like object.")

        self.out = out
        self.split_output = split_output
        self._files: dict[Hashable, IOBase] = {}

    def close(self) -> None:
        """Close all open files."""
        for f in self._files.values():
            f.close()

    def __enter__(self) -> "GribWriter":
        """Enter the runtime context related to this object.

        Returns
        -------
        GribWriter
            The GribWriter instance.
        """
        return self

    def __exit__(self, exc_type: type | None, exc_value: BaseException | None, trace: Any | None) -> None:
        """Exit the runtime context related to this object.

        Parameters
        ----------
        exc_type : Optional[type]
            The exception type.
        exc_value : Optional[BaseException]
            The exception value.
        trace : Optional[Any]
            The traceback object.
        """
        self.close()

    def write(
        self,
        *,
        values: Any | None,
        template: Any,
        metadata: dict[str, Any],
        check_nans: bool = False,
        missing_value: int | float = 9999,
    ) -> tuple:
        """Write a GRIB message to the target file.

        Parameters
        ----------
        values : Optional[Any]
            The values to encode.
        template : Any
            The template to use.
        metadata : Dict[str, Any]
            The metadata for the GRIB message.
        check_nans : bool, optional
            Whether to check for NaNs in the values.
        missing_value : Union[int, float], optional
            The value to use for missing data.

        Returns
        -------
        tuple
            The encoded GRIB handle and the file path.
        """

        while True:
            try:
                handle = encode_message(
                    values=values,
                    check_nans=check_nans,
                    metadata=metadata,
                    template=template,
                    missing_value=missing_value,
                )
                break
            except Exception as e:
                if metadata.get("edition") == 2:
                    raise
                # Try again with edition 2
                LOG.warning("Failed to encode GRIB message with edition 1, retrying with edition 2: %s", e)
                metadata["edition"] = 2

        file, path = self.target(handle)
        handle.write(file)

        return handle, path

    def target(self, handle: Any) -> tuple[IOBase, Path | str]:
        """Determine the target file for the GRIB message.

        Parameters
        ----------
        handle : Any
            The GRIB handle.

        Returns
        -------
        tuple
            The file object and the file path.
        """
        if self.split_output:
            assert not isinstance(self.out, IOBase), "Cannot split output when `out` is a file-like object."
            out = render_template(str(self.out), handle)
        else:
            out = self.out

        if isinstance(out, IOBase):
            # self._files['out'] = out
            return out, "out"

        if out not in self._files:
            self._files[out] = open(out, "wb")

        return self._files[out], out



# (C) Copyright 2025 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
from typing import TYPE_CHECKING

LOG = logging.getLogger(__name__)


class LazyModule:
    """Defer loading of a module until attribute access."""

    def __init__(self, module_name: str):
        self.__module_name = module_name
        self.__module = None

    def __getattr__(self, name: str):
        if self.__module is None:
            import importlib

            LOG.debug("Importing '%s'", self.__module_name)

            self.__module = importlib.import_module(self.__module_name)
        return getattr(self.__module, name)


# add heavy imports here, then they can be used in the rest of the codebase as regular imports
# with type checking and autocompletion: `from anemoi.inference.lazy import torch`
if TYPE_CHECKING:
    import torch as real_torch

torch: "real_torch" = LazyModule("torch")



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
import os
import pickle
import struct
from typing import Any

from anemoi.utils.logs import enable_logging_name
from anemoi.utils.logs import set_logging_name

from anemoi.inference.task import Task
from anemoi.inference.types import State

from ..transport import Transport
from . import transport_registry

LOG = logging.getLogger(__name__)


@transport_registry.register("processes")
class ProcessesTransport(Transport):
    """Transport implementation using processes."""

    def __init__(self, couplings: Any, tasks: dict[str, Any], *args: Any, **kwargs: Any) -> None:
        """Initialize the ProcessesTransport.

        Parameters
        ----------
        couplings : Any
            The couplings for the transport.
        tasks : Dict[str, Any]
            The tasks to be executed.
        """
        super().__init__(couplings, tasks)
        self.children: dict[int, str] = {}
        self.pipes: dict[tuple[str, str], tuple[int, int]] = {}
        enable_logging_name("main")

    def child_process(self, task: Any) -> int:
        """Run the task in a child process.

        Parameters
        ----------
        task : Any
            The task to be run.

        Returns
        -------
        int
            The exit status of the child process.
        """
        set_logging_name(task.name)

        # Close all the pipes that are not needed
        for (task1, task2), (read_fd, write_fd) in self.pipes.items():
            if task.name not in (task1, task2):
                os.close(read_fd)
                os.close(write_fd)

        code = 0
        try:
            task.run(self)
        except Exception as e:
            LOG.exception(e)
            code = 1
        LOG.info("Child process %s finished with code %d", task.name, code)
        return code

    def start(self) -> None:
        """Start the transport by forking processes for each task."""

        # Many to many pipes. May not scale well
        # We can use the couplings to reduce the number of pipes
        self.pipes = {}
        for task1 in self.tasks:
            for task2 in self.tasks:
                if task1 != task2:
                    read_fd, write_fd = os.pipe()
                    os.set_inheritable(read_fd, True)
                    os.set_inheritable(write_fd, True)
                    self.pipes[(task1, task2)] = (read_fd, write_fd)

        for name, task in self.tasks.items():
            pid = os.fork()
            if pid == 0:
                os._exit(self.child_process(task))
            else:
                self.children[pid] = name

        # We need to close the pipes in the parent process
        for read_fd, write_fd in self.pipes.values():
            os.close(read_fd)
            os.close(write_fd)

    def wait(self) -> None:
        """Wait for all child processes to complete and handle any errors."""
        while self.children:
            (pid, status) = os.wait()
            LOG.info(f"Child process {pid} ({self.children[pid]}) exited with status {status}")
            del self.children[pid]

            if status != 0:
                for pid in self.children:
                    os.kill(pid, 15)

            if status != 0:
                LOG.error("One of the child processes failed. Exiting.")
                raise RuntimeError(f"Child process {pid} ({self.children[pid]}) failed with status {status}")

    def send(self, sender: Task, target: Task, state: State, tag: int) -> None:
        """Send a state from the sender to the target.

        Parameters
        ----------
        sender : Any
            The task sending the state.
        target : Any
            The task receiving the state.
        state : Any
            The state to be sent.
        tag : int
            The tag associated with the state.
        """

        LOG.info(f"{sender}: sending to {target} [{tag}]")

        # TODO: something more efficient than pickle
        _, write_fd = self.pipes[(sender.name, target.name)]
        pickle_data = pickle.dumps(state)

        os.write(write_fd, struct.pack("!Q", tag))
        os.write(write_fd, struct.pack("!Q", len(pickle_data)))
        os.write(write_fd, pickle_data)

    def receive(self, receiver: Task, source: Task, tag: int) -> State:
        """Receive a state from the source to the receiver.

        Parameters
        ----------
        receiver : Any
            The task receiving the state.
        source : Any
            The task sending the state.
        tag : int
            The tag associated with the state.

        Returns
        -------
        Any
            The received state.
        """

        LOG.info(f"{receiver}: receiving from {source} [{tag}]")

        read_fd, _ = self.pipes[(source.name, receiver.name)]

        recieved_tag = struct.unpack("!Q", os.read(read_fd, 8))[0]
        assert recieved_tag == tag, (recieved_tag, tag)

        size = struct.unpack("!Q", os.read(read_fd, 8))[0]
        data = os.read(read_fd, size)
        state = pickle.loads(data)
        if isinstance(state, Exception):
            raise state

        return state



# (C) Copyright 2025 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from typing import Any

from anemoi.inference.context import Context
from anemoi.inference.state import combine_states
from anemoi.inference.types import Date
from anemoi.inference.types import State

from ..input import Input
from . import create_input
from . import input_registry

LOG = logging.getLogger(__name__)


@input_registry.register("split")
# @main_argument("splits")
class SplitInput(Input):

    trace_name = "split input"

    def __init__(self, context: Context, *splits, **kwargs: Any) -> None:

        all_variables = set(kwargs.get("variables"))
        assert all_variables is not None, "variables must be provided for split input"

        self.splits = {}

        default = None
        seen = set()

        for s in splits:
            assert isinstance(s, dict), "each split must be a dictionary"

            if "default" in s:
                default = s["default"]
                continue

            assert "source" in s, "each split must have a 'source' key"
            assert "variables" in s, "each split must have a 'variables' key"

            vars = s["variables"]

            if not isinstance(vars, list):
                vars = [vars]

            vars = set(vars)

            if not set(vars) <= all_variables:
                raise ValueError(
                    f"Variables {vars} not in the provided variables {all_variables} ({set(vars) - all_variables})"
                )

            self.splits[tuple(sorted(vars))] = create_input(
                context,
                s["source"],
                variables=vars,
                purpose=s.get("purpose"),
            )

            seen |= vars

        all_variables -= seen

        for i, vars1 in enumerate(splits):
            vars1 = set(vars1)
            for j, vars2 in enumerate(splits):
                vars2 = set(vars2)
                if i == j:
                    continue
                if not vars1.isdisjoint(vars2):
                    raise ValueError(f"Splits {vars1} and {vars2} overlap in variables {vars1 & vars2}")

        if all_variables:
            if default is None:
                raise ValueError(f"Variables {all_variables} not covered by any split and no default provided")

            LOG.debug(f"Variables {all_variables} not covered by any split, using default {default}")

            self.splits[tuple(sorted(all_variables))] = create_input(
                context,
                default,
                variables=sorted(all_variables),
                purpose=kwargs.get("purpose"),
            )

        assert len(self.splits) > 1, "At least two splits must be provided"

        self.splits = list(self.splits.values())

        super().__init__(context, **kwargs)

    def create_input_state(self, *, date: Date | None) -> State:
        """Create the input state for the repeated-dates input.

        Parameters
        ----------
        date : Date or None
            The date for the input state.

        Returns
        -------
        State
            The created input state.
        """

        # TODO: Consider caching the result
        states = [split.create_input_state(date=date) for split in self.splits]

        state = combine_states(*states)

        assert set(state["fields"]) == set(self.variables), (sorted(state["fields"]), sorted(self.variables))

        state["_input"] = self
        state["date"] = date

        return state

    def load_forcings_state(self, *, dates: list[Date], current_state: State) -> State:
        """Load the forcings state for repeated dates input.

        Parameters
        ----------
        dates : list of Date
            The list of dates for which to repeat the fields.
        current_state : State
            The current state to use for loading.

        Returns
        -------
        State
            The loaded and repeated forcings state.
        """
        assert len(dates) > 0, "dates must not be empty for repeated dates input"

        states = [split.load_forcings_state(dates=dates, current_state=current_state) for split in self.splits]
        state = combine_states(*states)

        assert set(state["fields"]) == set(self.variables)

        state["date"] = dates[-1]
        state["_input"] = self

        return state



# (C) Copyright 2024- Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import datetime

import numpy as np
import pytest

from anemoi.inference.config.run import RunConfiguration
from anemoi.inference.inputs.cutout import _mask_and_combine_states
from anemoi.inference.runner import Runner
from anemoi.inference.runners import create_runner
from anemoi.inference.testing import fake_checkpoints
from anemoi.inference.testing import files_for_tests


def test_mask_and_combine_states():
    states = [{"a": np.arange(5).astype(float)}, {"a": np.arange(5, 10).astype(float)}, {"a": np.arange(10, 15)}]
    masks = [np.zeros(5).astype(bool) for _ in states]
    masks[0][[1, 2]] = True
    masks[1][[2, 3]] = True
    masks[2][[2, 4]] = True

    combined_state: dict = {}
    for k in range(len(states)):
        mask = masks[k]
        new_state = states[k]
        combined_state = _mask_and_combine_states(combined_state, new_state, mask, ["a"])

    assert combined_state["a"].shape[0] == 6
    assert (
        combined_state["a"]
        == np.array(
            [
                states[0]["a"][1],
                states[0]["a"][2],
                states[1]["a"][2],
                states[1]["a"][3],
                states[2]["a"][2],
                states[2]["a"][4],
            ]
        )
    ).all()


@pytest.fixture
@fake_checkpoints
def runner() -> None:
    """Fake Runner for testing"""
    config = RunConfiguration.load(
        files_for_tests("unit/configs/simple.yaml"),
        overrides=dict(runner="testing", device="cpu", input="dummy", trace_path="trace.log"),
    )
    runner = create_runner(config)
    assert runner.checkpoint
    return runner


@fake_checkpoints
def test_cutout_no_mask(runner: Runner):
    from anemoi.inference.inputs.cutout import Cutout

    cutout_config = {
        "lam": {"mask": None, "dummy": {}},
        "global": {"mask": None, "dummy": {}},
    }
    cutout_input = Cutout(runner, variables=["2t"], **cutout_config)
    input_state = cutout_input.create_input_state(date=datetime.datetime.fromisoformat("2020-01-01T00:00"))
    number_of_grid_points = runner.checkpoint.number_of_grid_points

    assert "_mask" in input_state
    assert input_state["latitudes"].shape[0] == number_of_grid_points * 2

    assert all(input_state["_mask"]["lam"][slice(0, number_of_grid_points)])
    assert not all(input_state["_mask"]["lam"][slice(number_of_grid_points, None)])

    assert not all(input_state["_mask"]["global"][slice(0, number_of_grid_points)])
    assert all(input_state["_mask"]["global"][slice(number_of_grid_points, None)])


@fake_checkpoints
def test_cutout_with_slice(runner: Runner):
    from anemoi.inference.inputs.cutout import Cutout

    cutout_config = {
        "lam": {"mask": slice(0, 10), "dummy": {}},
        "global": {"mask": slice(10, 25), "dummy": {}},
    }
    cutout_input = Cutout(runner, variables=["2t"], **cutout_config)
    assert list(cutout_input.sources.keys()) == ["lam", "global"]

    input_state = cutout_input.create_input_state(date=datetime.datetime.fromisoformat("2020-01-01T00:00"))

    assert "_mask" in input_state
    assert input_state["latitudes"].shape[0] == 25

    assert all(input_state["_mask"]["lam"][slice(0, 10)])
    assert not all(input_state["_mask"]["lam"][slice(10, None)])

    assert not all(input_state["_mask"]["global"][slice(0, 10)])
    assert all(input_state["_mask"]["global"][slice(10, None)])


@fake_checkpoints
def test_cutout_with_array(runner: Runner):
    from anemoi.inference.inputs.cutout import Cutout

    number_of_grid_points = runner.checkpoint.number_of_grid_points

    lam_mask = np.zeros(number_of_grid_points, dtype=bool)
    lam_mask[:10] = True

    global_mask = np.zeros(number_of_grid_points, dtype=bool)
    global_mask[10:25] = True

    cutout_config = {
        "lam": {"mask": lam_mask, "dummy": {}},
        "global": {"mask": global_mask, "dummy": {}},
    }
    cutout_input = Cutout(runner, variables=["2t"], **cutout_config)
    input_state = cutout_input.create_input_state(date=datetime.datetime.fromisoformat("2020-01-01T00:00"))

    assert "_mask" in input_state
    assert input_state["latitudes"].shape[0] == 25

    assert all(input_state["_mask"]["lam"][slice(0, 10)])
    assert not all(input_state["_mask"]["lam"][slice(10, None)])

    assert not all(input_state["_mask"]["global"][slice(0, 10)])
    assert all(input_state["_mask"]["global"][slice(15, None)])



# (C) Copyright 2025 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


from datetime import datetime
from datetime import timedelta

import numpy as np
import pytest

pytest_plugins = "anemoi.utils.testing"


STATE_NPOINTS = 50


@pytest.fixture
def state():
    """Fixture to create a mock state for testing."""

    return {
        "latitudes": np.random.uniform(-90, 90, size=STATE_NPOINTS),
        "longitudes": np.random.uniform(-180, 180, size=STATE_NPOINTS),
        "fields": {
            "2t": np.random.uniform(250, 310, size=STATE_NPOINTS),
            "z_850": np.random.uniform(500, 1500, size=STATE_NPOINTS),
        },
        "date": datetime(2020, 1, 1, 0, 0),
        "step": timedelta(hours=6),
    }


@pytest.fixture
def extract_mask_npy(tmp_path):
    """Fixture to create a mock mask in .npy format."""
    mask = np.random.choice([True, False], size=STATE_NPOINTS)
    mask_path = tmp_path / "extract_mask.npy"
    np.save(mask_path, mask)
    return str(mask_path)


@pytest.fixture
def assign_mask_npy(tmp_path):
    """Fixture to create a mock mask in .npy format."""
    mask = np.zeros(STATE_NPOINTS + 10, dtype=bool)
    mask[:STATE_NPOINTS] = True
    mask_path = tmp_path / "assign_mask.npy"
    np.save(mask_path, mask)
    return str(mask_path)



# (C) Copyright 2025 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from collections import defaultdict
from collections.abc import Iterable

import numpy as np

from anemoi.inference.input import Input
from anemoi.inference.inputs import create_input
from anemoi.inference.types import Date
from anemoi.inference.types import ProcessorConfig
from anemoi.inference.types import State

from . import input_registry

LOG = logging.getLogger(__name__)


def _mask_and_combine_states(
    existing_state: State,
    new_state: State,
    mask: np.ndarray | slice,
    fields: Iterable[str],
) -> State:
    """Mask and combine two states.

    Existing can be None, in which case new_state masked is returned.

    Parameters
    ----------
    existing_state : State
        The state to be combined into.
    new_state : State
        The other state to combine.
    mask : np.ndarray
        The mask to apply to new_state.
    fields: Iterable[str]
        The fields to combine in the states

    Returns
    -------
    State
        The combined state
    """
    was_empty = existing_state == {}

    for field in fields:
        if was_empty:
            existing_state[field] = new_state[field][..., mask]
        else:
            existing_state[field] = np.concatenate(
                [existing_state[field], new_state[field][..., mask]],
                axis=-1,
            )

    return existing_state


def _realise_mask(sli: slice | np.ndarray, state: dict) -> np.ndarray:
    """Realise a slice or array into an bool array based on the state shape."""
    return np.ones(len(state["latitudes"]), dtype=bool)[sli]


def _extract_and_add_private_attributes(
    private_attributes: defaultdict[str, dict], state: State, name: str
) -> defaultdict[str, dict]:
    """Extract and add private attributes to the state.

    Will nest the attributes under the name provided, maintaining the original key.

    Parameters
    ----------
    private_attributes : defaultdict[str, dict]
        The dictionary to contain private attributes.
    state : State
        The state to which the attributes will be retrieved.
    name : str
        The name of the sub-state to record the attributes under.

    Returns
    -------
    defaultdict[str, dict]
        The updated private attributes dictionary.
    """

    for key in (k for k in state.keys() if k.startswith("_")):
        private_attributes[key][name] = state[key]
    return private_attributes


@input_registry.register("cutout")
class Cutout(Input):
    """Combines one or more LAMs into a global source using cutouts."""

    # TODO: Does this need an ordering?

    def __init__(
        self,
        context,
        variables: list[str] | None,
        pre_processors: list[ProcessorConfig] | None = None,
        purpose: str | None = None,
        **sources: dict[str, dict],
    ):
        """Create a cutout input from a list of sources.

        Parameters
        ----------
        context : dict
            The context runner.
        sources : dict of sources
            A dictionary of sources to combine.
        """

        super().__init__(context, variables=variables, pre_processors=pre_processors, purpose=purpose)

        if pre_processors:
            raise ValueError(
                "Input `cutout` does not support pre_processors. Please add pre_processors to each of the sub-input."
            )

        self.sources: dict[str, Input] = {}
        self.masks: dict[str, np.ndarray | slice] = {}

        for src, cfg in sources.items():
            if isinstance(cfg, str):
                mask = f"{src}/cutout_mask"
            else:
                cfg = cfg.copy()
                mask = cfg.pop("mask", f"{src}/cutout_mask")

            self.sources[src] = create_input(
                context, cfg, variables=variables, pre_processors=pre_processors, purpose=purpose
            )

            if isinstance(mask, str):
                self.masks[src] = self.sources[src].checkpoint.load_supporting_array(mask)
            else:
                self.masks[src] = mask if mask is not None else slice(0, None)  # type: ignore

    def __repr__(self):
        """Return a string representation of the Cutout object."""
        return f"Cutout({self.sources})"

    def create_input_state(self, *, date: Date | None) -> State:
        """Create the input state for the given date.

        Parameters
        ----------
        date : Optional[Date]
            The date for which to create the input state.

        Returns
        -------
        State
            The created input state.
        """

        LOG.info(f"Concatenating states from {self.sources}")

        _mask_private_attributes = {}
        _private_attributes = defaultdict(dict)

        combined_state = {}

        for source in self.sources.keys():
            source_state = self.sources[source].create_input_state(date=date)
            source_mask = self.masks[source]

            # Create the mask front padded with zeros
            # to match the length of the combined state
            _realised_mask = _realise_mask(source_mask, source_state)
            current_length = len(combined_state.get("latitudes", []))

            _mask_private_attributes[source] = np.concatenate(
                (
                    np.zeros((current_length,), dtype=bool),
                    _realised_mask,
                ),
                axis=-1,
            )
            # Combine the private attributes
            _private_attributes = _extract_and_add_private_attributes(_private_attributes, source_state, source)

            combined_state = _mask_and_combine_states(
                combined_state, source_state, source_mask, ["longitudes", "latitudes"]
            )
            combined_state["fields"] = _mask_and_combine_states(
                combined_state.get("fields", {}), source_state["fields"], source_mask, source_state["fields"].keys()
            )

            for key in (k for k in source_state.keys() if not k.startswith("_")):
                combined_state.setdefault(key, source_state[key])

        # Pad the masks to the total length of the combined state
        # then add them to the private attributes
        total_length = len(combined_state["latitudes"])
        for sub_mask in _mask_private_attributes:
            mask = _mask_private_attributes[sub_mask]
            _mask_private_attributes[sub_mask] = np.pad(mask, (0, total_length - len(mask)), constant_values=False)

        _private_attributes["_mask"] = _mask_private_attributes

        combined_state.update(_private_attributes)
        return combined_state

    def load_forcings_state(self, *, dates: list[Date], current_state: State) -> State:
        """Load the forcings state for the given variables and dates.

        Parameters
        ----------

        dates : List[Date]
            List of dates for which to load the forcings.
        current_state : State
            The current state of the input.

        Returns
        -------
        State
            The loaded forcings state.
        """

        combined_fields = {}

        for source in self.sources.keys():
            source_state = self.sources[source].load_forcings_state(dates=dates, current_state=current_state)["fields"]
            source_mask = self.masks[source]

            combined_fields = _mask_and_combine_states(combined_fields, source_state, source_mask, source_state.keys())

        current_state["fields"] |= combined_fields
        return current_state



# (C) Copyright 2025 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import datetime
import itertools

from anemoi.inference.types import State


def check_state(state: State, title: str = "<state>") -> None:
    """Check the state for consistency.

    Parameters
    ----------
    state : dict
        The state to check.
    title : str
        The title of the state (for logging).

    Raises
    ------
    ValueError
        If the state is not consistent.
    """

    if not isinstance(state, dict):
        raise ValueError(f"State '{title}' is not a dictionary: {state}")

    input = state.get("_input")

    if "fields" not in state:
        raise ValueError(f"State '{title}' does not contain 'fields': {state} ({input=})")

    shape = None

    for field, values in state["fields"].items():
        if shape is None:
            shape = values.shape
        elif shape != values.shape:
            raise ValueError(
                f"Field '{field}' in state '{title}' has different shape: " f"{shape} and {values.shape} ({input=})."
            )

    date = state.get("date")
    if date is None and len(state["fields"]) > 0:
        # date can be None for an empty input
        if not isinstance(date, datetime.datetime):
            raise ValueError(f"State '{title}' does not contain 'date', or it is not a datetime: {date} ({input=})")


def combine_states(*states: State) -> State:
    """Combine multiple states into one.

    Parameters
    ----------
    states : list
        The states to combine.

    Returns
    -------
    dict
        The combined state.
    """
    import numpy as np

    assert len(states) > 1, "At least two states must be provided"

    combined = states[0].copy()
    combined["fields"] = combined["fields"].copy()
    shape = None
    first_input = combined.get("_input")

    for state in states[1:]:

        this_input = state.get("_input")

        for name, values in itertools.chain(combined["fields"].items(), state.get("fields", {}).items()):
            if shape is None:
                shape = values.shape
            elif shape != values.shape:
                raise ValueError(
                    f"Field '{name}' has different shape in the states: "
                    f"{shape} and {values.shape}."
                    f" Input: {first_input} vs {this_input}."
                )

        if not set(combined["fields"]).isdisjoint(state["fields"]):
            raise ValueError(
                f"Some states have overlapping fields:"
                f" {set(combined['fields']).intersection(state['fields'])}"
                f" Input: {first_input} vs {this_input}."
            )

        combined["fields"].update(state.get("fields", {}))
        for key, value in state.items():
            if key == "fields":
                continue

            if key.startswith("_"):
                continue

            if combined.get(key) is None:
                combined[key] = value
                continue

            if value is None:
                continue

            if type(combined[key]) is not type(value):
                raise ValueError(
                    f"Key '{key}' has different types in the states: " f"{type(combined[key])} and {type(value)}."
                )

            if isinstance(value, np.ndarray) and isinstance(combined[key], np.ndarray):
                if not np.array_equal(combined[key], value):
                    raise ValueError(
                        f"Key '{key}' has different array values in the states: "
                        f"{combined[key]} and {value}."
                        f" Input: {first_input} vs {this_input}."
                    )
                continue

            if combined[key] != value:
                raise ValueError(
                    f"Key '{key}' has different values in the states: "
                    f"{combined[key]} and {value} ({shape})."
                    f" Input: {first_input} vs {this_input}."
                )

    return combined


def reduce_state(state: State) -> State:
    """Create a new state which is a projection of the original state on the last step in the multi-steps dimension.

    Parameters
    ----------
    state : State
        The original state.

    Returns
    -------
    State
        The reduced state.
    """
    reduced_state = state.copy()
    reduced_state["fields"] = {}
    for field, values in state["fields"].items():
        if len(values.shape) > 1:
            reduced_state["fields"][field] = values[-1, :]
        else:
            reduced_state["fields"][field] = values
    return reduced_state



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import base64
import logging
import os
import zlib
from typing import Any

import earthkit.data as ekd

from . import IndexTemplateProvider
from . import template_provider_registry

LOG = logging.getLogger(__name__)

# 1 - Get a GRIB with mars: retrieve,param=tp,levtype=sfc,type=fc,step=6,target=data.grib,grid=0.25/0.25
# 2 - grib_set -s edition=2,packingType=grid_ccsds data.grib data.grib2
# 3 - grib_set -d 0 data.grib2 out.grib
# 4 - python -c 'import base64, sys, zlib;print(base64.b64encode(zlib.compress(open(sys.argv[1], "rb").read())))' out.grib

# tp in grib2, 0.25/0.25 grid


@template_provider_registry.register("builtin")
class BuiltinTemplates(IndexTemplateProvider):
    """Builtin templates provider."""

    def __init__(self, manager: Any, index_path: str | None = None) -> None:
        """Initialize the BuiltinTemplates instance.

        Parameters
        ----------
        manager : Any
            The manager instance.
        index_path : Optional[str], optional
            The path to the index file, by default None.
        """
        if index_path is None:
            index_path = os.path.join(os.path.dirname(__file__), "builtin.yaml")

        super().__init__(manager, index_path)

    def load_template(self, grib: str, lookup: dict[str, Any]) -> ekd.Field | None:
        """Load the template for the given GRIB and lookup.

        Parameters
        ----------
        grib : str
            The GRIB string.
        lookup : Dict[str, Any]
            The lookup dictionary.

        Returns
        -------
        Optional[ekd.Field]
            The loaded template field if found, otherwise None.
        """
        import earthkit.data as ekd

        template = zlib.decompress(base64.b64decode(grib))
        return ekd.from_source("memory", template)[0]



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from collections.abc import Generator
from contextlib import contextmanager
from functools import cached_property
from typing import Any

from .protocol import MetadataProtocol

LOG = logging.getLogger(__name__)


@contextmanager
def patch_function(target: Any, attribute: str, replacement: Any) -> Generator[None, None, None]:
    """Context manager to temporarily replace an attribute of a target object.

    Parameters
    ----------
    target : object
        The target object whose attribute will be replaced.
    attribute : str
        The name of the attribute to replace.
    replacement : any
        The replacement value for the attribute.

    Returns
    -------
    Generator[None, None, None]
        The context manager.
    """
    original = getattr(target, attribute)
    setattr(target, attribute, replacement)
    try:
        yield
    finally:
        setattr(target, attribute, original)


class PatchMixin(MetadataProtocol):

    # `self` is a `Metadata` object

    def patch_metadata(
        self,
        supporting_arrays: dict[str, Any],
        root: str,
    ) -> tuple[dict[str, Any], dict[str, Any]]:
        """Patch the metadata with supporting arrays and root.

        Parameters
        ----------
        supporting_arrays : dict
            The supporting arrays to patch.
        root : str
            The root path for the supporting arrays.

        Returns
        -------
        tuple
            The patched metadata and supporting arrays.
        """

        metadata, supporting_arrays = self._from_zarr
        self._metadata["dataset"] = metadata
        self._supporting_arrays = supporting_arrays
        return self._metadata, self._supporting_arrays

    @cached_property
    def _from_zarr(self) -> tuple[dict[str, Any], dict[str, Any]]:
        """Open the dataset and fetch metadata and supporting arrays."""
        # We assume that the datasets are reachable via the content of
        # ~/.config/anemoi/settings.toml.

        ds = self.open_dataset()
        return ds.metadata(), ds.supporting_arrays()



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from argparse import ArgumentParser
from argparse import Namespace
from copy import deepcopy

import numpy as np

from . import Command

LOG = logging.getLogger(__name__)


def diff(old, new, *path):

    result = False

    if isinstance(old, dict) and isinstance(new, dict):
        keys1 = set(old.keys()) if isinstance(old, dict) else set()
        keys2 = set(new.keys()) if isinstance(new, dict) else set()

        for key in keys1.union(keys2):
            if key in old and key in new:
                result = diff(old[key], new[key], *path, key) or result
            elif key in old:
                LOG.info(f"{'.'.join(path + (key,))} removed")
                result = True
            elif key in new:
                LOG.info(f"{'.'.join(path + (key,))} added")
                result = True

        return result

    if isinstance(old, list) and isinstance(new, list):
        if len(old) != len(new):
            LOG.info(f"Difference at {path}: List lengths differ ({len(old)} != {len(new)})")
            result = True
        for i, (item1, item2) in enumerate(zip(old, new)):
            result = diff(item1, item2, *path, str(i)) or result
        return result

    if isinstance(old, np.ndarray) and isinstance(new, np.ndarray):
        if not np.array_equal(old, new):
            LOG.info(f"Difference at {path}: Array values differ")
            result = True
        return result

    if old != new:
        LOG.info(f"{'.'.join(path)}: {old} != {new}")
        result = True

    return result


class PatchCmd(Command):
    """Patch a checkpoint file."""

    def add_arguments(self, command_parser: ArgumentParser) -> None:
        """Add arguments to the command parser.

        Parameters
        ----------
        command_parser : ArgumentParser
            The argument parser to which the arguments will be added.
        """
        command_parser.add_argument("path", help="Path to the checkpoint.")
        # command_parser.add_argument("--sanitise", action="store_true", help="Sanitise the metadata.")

    def run(self, args: Namespace) -> None:
        """Run the patch command.

        Parameters
        ----------
        args : Namespace
            The arguments passed to the command.
        """
        from anemoi.utils.checkpoints import load_metadata
        from anemoi.utils.checkpoints import metadata_root
        from anemoi.utils.checkpoints import replace_metadata

        from anemoi.inference.metadata import Metadata

        root = metadata_root(args.path)

        original_metadata, original_supporting_arrays = load_metadata(args.path, supporting_arrays=True)
        original_metadata = deepcopy(original_metadata)
        original_supporting_arrays = deepcopy(original_supporting_arrays)

        metadata, supporting_arrays = Metadata(original_metadata).patch_metadata(original_supporting_arrays, root)

        if diff(metadata, original_metadata) or diff(original_supporting_arrays, supporting_arrays):
            LOG.info("Patching metadata")
            assert "sources" in metadata["dataset"]
            replace_metadata(args.path, metadata, supporting_arrays)


command = PatchCmd



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


"""List of precisions supported by the inference runner."""


from functools import cached_property


class LazyDict:
    """A dictionary that lazily loads its values.
    So we don't import torch at the top level, which can be slow.
    """

    def get(self, key, default=None):
        return self._mapping.get(key, default)

    def keys(self):
        return self._mapping.keys()

    def values(self):
        return self._mapping.values()

    def items(self):
        return self._mapping.items()

    def __getitem__(self, key):
        return self._mapping[key]

    @cached_property
    def _mapping(self):
        import torch

        return {
            "16-mixed": torch.float16,
            "16": torch.float16,
            "32": torch.float32,
            "b16-mixed": torch.bfloat16,
            "b16": torch.bfloat16,
            "bf16-mixed": torch.bfloat16,
            "bf16": torch.bfloat16,
            "bfloat16": torch.bfloat16,
            "f16": torch.float16,
            "f32": torch.float32,
            "float16": torch.float16,
            "float32": torch.float32,
        }


PRECISIONS = LazyDict()

if __name__ == "__main__":
    # This is just to make sure that the module can be imported without errors.
    # It will not be executed when the module is imported, only when run as a script.

    print("Available precisions:", list(PRECISIONS.keys()))
    print("Available precisions:", list(PRECISIONS.values()))
    print("Available precisions:", list(PRECISIONS.items()))

    print("Torch float16:", PRECISIONS["16"])
    print("Torch bfloat16:", PRECISIONS["b16"])
    print("Torch float32:", PRECISIONS["32"])



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from __future__ import annotations

import logging
import shutil
from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any
from typing import Literal

import numpy as np

from anemoi.inference.context import Context
from anemoi.inference.state import reduce_state
from anemoi.inference.types import State

from ..decorators import main_argument
from ..output import Output
from . import output_registry

LOG = logging.getLogger(__name__)

if TYPE_CHECKING:
    from zarr.storage import StoreLike


def create_zarr_array(
    store: "StoreLike",
    name: str,
    shape: tuple,
    dtype: str,
    dimensions: tuple[str, ...],
    chunks: tuple[int, ...] | str | bool,
    fill_value: float | None = None,
) -> Any:
    """Create a Zarr array with the given parameters.

    Parses the Zarr version to handle differences in API between versions 2 and 3.
    """
    import zarr

    chunks = chunks if zarr.__version__ >= "3" else chunks if not chunks == "auto" else True

    if zarr.__version__ >= "3":
        from zarr import create_array
    else:

        def create_array(*, store, **kwargs):
            """Create a Zarr array using the zarr 2 API."""
            return store.create_dataset(**kwargs)

    array = create_array(
        store=store,
        name=name,
        shape=shape,
        dtype=dtype,
        chunks=chunks,
        fill_value=fill_value,
        dimension_names=dimensions,
        overwrite=True,
    )
    array.attrs["_ARRAY_DIMENSIONS"] = list(dimensions)

    return array


@output_registry.register("zarr")  # type: ignore
@main_argument("store")
class ZarrOutput(Output):
    """Zarr output class."""

    def __init__(
        self,
        context: Context,
        store: "StoreLike",
        variables: list[str] | None = None,
        output_frequency: int | None = None,
        write_initial_state: bool | None = None,
        missing_value: float | None = np.nan,
        float_size: str = "f4",
        chunks: tuple[int, ...] | Literal["auto"] | bool = "auto",
    ) -> None:
        """Initialise the ZarrOutput object.

        Parameters
        ----------
        context : dict
            The context dictionary.
        store : StoreLike
            The Zarr store to save the data.
            Can be a file path or a Zarr store.
            If an existing store is provided, it is assumed to
            be a writable store and empty.
        variables : list, optional
            The list of variables to write, by default None.
        output_frequency : int, optional
            The frequency of output, by default None.
        write_initial_state : bool, optional
            Whether to write the initial state, by default None.
        float_size : str, optional
            The size of the float, by default "f4".
        missing_value : float, optional
            The missing value, by default np.nan.
        chunks : tuple[int, ...] | Literal['auto'] | bool, optional
            The chunk size for the Zarr arrays, by default 'auto'.
        """

        super().__init__(
            context, variables=variables, output_frequency=output_frequency, write_initial_state=write_initial_state
        )

        self.zarr_store: StoreLike = store
        self.missing_value = missing_value
        self.chunks = chunks
        self.float_size = float_size

        self._vars = {}

    def __repr__(self) -> str:
        """Return a string representation of the ZarrOutput object."""
        return f"ZarrOutput({self.zarr_store})"

    def open(self, state: State) -> None:
        """Open the Zarr file and initialize dimensions.

        Parameters
        ----------
        state : State
            The state dictionary.
        """
        import zarr

        if isinstance(self.zarr_store, (str, Path)):
            zarr_store = Path(self.zarr_store)
            zarr_store.parent.mkdir(parents=True, exist_ok=True)

            if zarr_store.exists():
                LOG.warning(f"Zarr store {self.zarr_store} already exists. It will be overwritten.")
                shutil.rmtree(self.zarr_store)

            if zarr.__version__ >= "3":
                from zarr.storage import LocalStore

                self.zarr_store = LocalStore(self.zarr_store)
            else:
                from zarr.storage import DirectoryStore

                self.zarr_store = DirectoryStore(self.zarr_store)

        if zarr.__version__ >= "3":
            self.zarr_group = self.zarr_store
        else:
            self.zarr_group = zarr.open_group(self.zarr_store, mode="w")

        state = self.post_process(state)  # make sure state is post-processed
        values = len(state["latitudes"])

        time = 0

        self.reference_date = state["date"]

        lead_time = getattr(self.context, "lead_time", None)
        time_step = self.context.checkpoint.timestep

        if lead_time is None:
            raise RuntimeError(
                "When setting up the ZarrOutput, the `lead_time` was not yet set on the context, therefore unable to construct the arrays."
            )

        time = lead_time // time_step
        time += int(self.write_step_zero)

        if reference_date := getattr(self.context, "reference_date", None):
            self.reference_date = reference_date

        if not self.write_step_zero and time_step is not None:
            self.reference_date -= time_step

        self.time_size = time
        self.time_array = create_zarr_array(
            self.zarr_group,
            name="time",
            shape=(self.time_size,),
            dtype="i4",
            dimensions=("time",),
            chunks=(1,),
        )
        self.time_array.attrs.update(
            {
                "units": f"seconds since {self.reference_date}",
                "calendar": "gregorian",
            }
        )

        latitudes = state["latitudes"]
        self.latitude_var = create_zarr_array(
            self.zarr_group,
            name="latitude",
            shape=(values,),
            dtype=self.float_size,
            dimensions=("latitude",),
            chunks=self.chunks,
            fill_value=self.missing_value,
        )
        self.latitude_var.attrs.update({"units": "degrees_north", "long_name": "latitude"})

        longitudes = state["longitudes"]
        self.longitude_var = create_zarr_array(
            self.zarr_group,
            name="longitude",
            shape=(values,),
            dtype=self.float_size,
            dimensions=("longitude",),
            chunks=self.chunks,
            fill_value=self.missing_value,
        )

        self.n = 0
        self.latitude_var[:] = latitudes
        self.longitude_var[:] = longitudes

    def _variable_array(self, name: str, values_size: int) -> Any:
        """Get the variable array by name.

        Creates the variable array if it does not exist.

        Parameters
        ----------
        name : str
            The name of the variable.
        values_size : int
            The size of the values dimension.

        Returns
        -------
        Any
            The variable array.
        """
        if name in self._vars:
            return self._vars[name]

        self._vars[name] = create_zarr_array(
            self.zarr_group,
            name=name,
            shape=(self.time_size, values_size),
            dtype=self.float_size,
            dimensions=("time", "values"),
            chunks=self.chunks,
            fill_value=self.missing_value,
        )

        variable = self.typed_variables[name]
        self._vars[name].attrs.update({"param": variable.grib_keys.get("param", name)})
        return self._vars[name]

    def write_initial_state(self, state: State) -> None:
        """Write the initial state to the Zarr file."""
        state = reduce_state(state)  # Reduce to only the last step

        return super().write_initial_state(state)

    def write_step(self, state: State) -> None:
        """Write the state.

        Parameters
        ----------
        state : State
            The state dictionary.
        """
        step = state["date"] - self.reference_date
        self.time_array[self.n] = step.total_seconds()

        values = len(state["latitudes"])

        for name, value in state["fields"].items():
            if self.skip_variable(name):
                continue
            self._variable_array(name, values)[self.n] = value

        self.n += 1

    def close(self) -> None:
        """Close the Zarr file."""
        import zarr

        if zarr.__version__ >= "3":
            from zarr.abc.store import Store
        else:
            from zarr.storage import BaseStore as Store

        if self.zarr_store is not None and not isinstance(self.zarr_store, str):
            if isinstance(self.zarr_store, Store):
                zarr.consolidate_metadata(self.zarr_store)
                self.zarr_store.close()
            self.zarr_store = None



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
from pathlib import Path

import numpy as np
from anemoi.utils.grib import units

from anemoi.inference.context import Context
from anemoi.inference.decorators import ensure_dir
from anemoi.inference.decorators import main_argument
from anemoi.inference.types import FloatArray
from anemoi.inference.types import ProcessorConfig
from anemoi.inference.types import State
from anemoi.inference.utils.templating import render_template

from ..output import Output
from . import output_registry

LOG = logging.getLogger(__name__)


def fix(lons: FloatArray) -> FloatArray:
    """Fix longitudes greater than 180 degrees.

    Parameters
    ----------
    lons : FloatArray
        Array of longitudes.

    Returns
    -------
    FloatArray
        Fixed array of longitudes.
    """
    return np.where(lons > 180, lons - 360, lons)


@output_registry.register("plot")
@main_argument("dir")
@ensure_dir("dir")
class PlotOutput(Output):
    """Use `earthkit-plots` to plot the outputs."""

    def __init__(
        self,
        context: Context,
        dir: Path,
        *,
        variables: list[str] | None = None,
        mode: str = "subplots",
        domain: str | list[str] | None = None,
        template: str = "plot_{date}.{format}",
        format: str = "png",
        post_processors: list[ProcessorConfig] | None = None,
        output_frequency: int | None = None,
        write_initial_state: bool | None = None,
        **kwargs,
    ) -> None:
        """Initialise the PlotOutput.

        Parameters
        ----------
        context : Context
            The context.
        dir : Path
            The directory to save the plots.
            If the directory does not exist, it will be created.
        variables : list, optional
            The list of variables to plot, by default all.
        mode : str, optional
            The plotting mode, can be "subplots" or "overlay", by default "subplots".
        domain : str | list[str] | None, optional
            The domain/s to plot, by default None.
        template : str, optional
            The template for plot filenames, by default "plot_{date}.{format}".
            Has access to the following variables:
            - date: the date of the forecast step
            - basetime: the base time of the forecast
            - domain: the domain being plotted
            - format: the format of the plot
            - variables: the variables being plotted (joined by underscores)
        format : str, optional
            The format of the plot, by default "png".
        post_processors : Optional[List[ProcessorConfig]], default None
            Post-processors to apply to the input
        output_frequency : int, optional
            The frequency of output, by default None.
        write_initial_state : bool, optional
            Whether to write the initial state, by default None.
        **kwargs : Any
            Additional keyword arguments to pass to `earthkit.plots.quickplot`.
        """

        super().__init__(
            context,
            variables=variables,
            post_processors=post_processors,
            output_frequency=output_frequency,
            write_initial_state=write_initial_state,
        )

        self.dir = dir
        self.format = format
        self.variables = variables
        self.template = template
        self.domain = domain
        self.mode = mode
        self.kwargs = kwargs

    def write_step(self, state: State) -> None:
        """Write a step of the state.

        Parameters
        ----------
        state : State
            The state dictionary.
        """
        import earthkit.data as ekd
        import earthkit.plots as ekp

        longitudes = fix(state["longitudes"])
        latitudes = state["latitudes"]
        date = state["date"]
        basetime = date - state["step"]

        plotting_fields = []

        for name, values in state["fields"].items():
            if self.skip_variable(name):
                continue

            variable = self.typed_variables[name]
            param = variable.param

            plotting_fields.append(
                ekd.ArrayField(
                    values,
                    {
                        "param": param,
                        "shortName": param,
                        "variable_name": param,
                        "step": state["step"],
                        "base_datetime": basetime,
                        "latitudes": latitudes,
                        "longitudes": longitudes,
                        "units": units(param),
                    },
                )
            )

        fig = ekp.quickplot(
            ekd.FieldList.from_fields(plotting_fields), mode=self.mode, domain=self.domain, **self.kwargs
        )
        fname = render_template(
            self.template,
            {
                "date": date,
                "basetime": basetime,
                "domain": self.domain,
                "format": self.format,
                "variables": "_".join(self.variables or []),
            },
        )
        fname = self.dir / fname

        fig.save(fname)
        del fig



# (C) Copyright 2025 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


from typing import Any


class Variables:

    def __init__(self, runner: Any) -> None:
        """Initialise the Variables object.

        Parameters
        ----------
        runner : Any
            The runner object containing the checkpoint.
        """
        self.runner = runner
        self.checkpoint = runner.checkpoint

    ###############################################################################
    @classmethod
    def default_runner_input_variables_include_exclude(cls) -> dict[str, list[str]]:
        """Get include/exclude lists for default runner input variables.

        Returns
        -------
        dict
            Dictionary with 'include' and 'exclude' lists for default input variables.
        """
        return dict(
            include=["prognostic", "forcing"],
            exclude=["computed", "diagnostic"],
        )

    def default_input_variables(self) -> Any:
        """Select default input variables from the checkpoint.

        Returns
        -------
        Any
            The selected variables.
        """
        return self.checkpoint.select_variables(
            **self.default_runner_input_variables_include_exclude(),
        )

    def default_input_variables_and_mask(self) -> Any:
        """Select default input variables and masks from the checkpoint.

        Returns
        -------
        Any
            The selected variables and masks.
        """
        return self.checkpoint.select_variables_and_masks(
            **self.default_runner_input_variables_include_exclude(),
        )

    ###############################################################################
    @classmethod
    def retrieved_constant_forcings_variables_include_exclude(cls) -> dict[str, list[str]]:
        """Get include/exclude lists for retrieved constant forcings variables.

        Returns
        -------
        dict
            Dictionary with 'include' and 'exclude' lists for constant forcings variables.
        """
        return dict(
            include=["constant+forcing"],
            exclude=["computed", "diagnostic", "prognostic"],
        )

    def retrieved_constant_forcings_variables(self) -> Any:
        """Select retrieved constant forcings variables from the checkpoint.

        Returns
        -------
        Any
            The selected constant forcings variables.
        """
        return self.checkpoint.select_variables(
            **self.retrieved_constant_forcings_variables_include_exclude(),
        )

    def retrieved_constant_forcings_variables_and_mask(self) -> Any:
        """Select retrieved constant forcings variables and masks from the checkpoint.

        Returns
        -------
        Any
            The selected constant forcings variables and masks.
        """
        return self.checkpoint.select_variables_and_masks(
            **self.retrieved_constant_forcings_variables_include_exclude(),
        )

    ###############################################################################
    @classmethod
    def retrieved_prognostic_variables_include_exclude(cls) -> dict[str, list[str]]:
        """Get include/exclude lists for retrieved prognostic variables.

        Returns
        -------
        dict
            Dictionary with 'include' and 'exclude' lists for prognostic variables.
        """
        return dict(
            include=["prognostic"],
            exclude=["computed", "diagnostic", "forcing"],
        )

    def retrieved_prognostic_variables(self) -> Any:
        """Select retrieved prognostic variables from the checkpoint.

        Returns
        -------
        Any
            The selected prognostic variables.
        """
        return self.checkpoint.select_variables(
            **self.retrieved_prognostic_variables_include_exclude(),
        )

    def retrieved_prognostic_variables_and_mask(self) -> Any:
        """Select retrieved prognostic variables and masks from the checkpoint.

        Returns
        -------
        Any
            The selected prognostic variables and masks.
        """
        return self.checkpoint.select_variables_and_masks(
            **self.retrieved_prognostic_variables_include_exclude(),
        )

    ###############################################################################
    @classmethod
    def computed_constant_forcings_variables_include_exclude(cls) -> dict[str, list[str]]:
        """Get include/exclude lists for computed constant forcings variables.

        Returns
        -------
        dict
            Dictionary with 'include' list for computed constant forcings variables.
        """
        return dict(
            include=["computed+constant+forcing"],
        )

    def computed_constant_forcings_variables(self) -> Any:
        """Select computed constant forcings variables from the checkpoint.

        Returns
        -------
        Any
            The selected computed constant forcings variables.
        """
        return self.checkpoint.select_variables(
            **self.computed_constant_forcings_variables_include_exclude(),
        )

    def computed_constant_forcings_variables_and_mask(self) -> Any:
        """Select computed constant forcings variables and masks from the checkpoint.

        Returns
        -------
        Any
            The selected computed constant forcings variables and masks.
        """
        return self.checkpoint.select_variables_and_masks(
            **self.computed_constant_forcings_variables_include_exclude(),
        )

    ###############################################################################
    @classmethod
    def retrieved_dynamic_forcings_variables_include_exclude(cls) -> dict[str, list[str]]:
        """Get include/exclude lists for retrieved dynamic forcings variables.

        Returns
        -------
        dict
            Dictionary with 'include' and 'exclude' lists for dynamic forcings variables.
        """
        return dict(
            include=["forcing"],
            exclude=["computed", "constant"],
        )

    def retrieved_dynamic_forcings_variables(self) -> Any:
        """Select retrieved dynamic forcings variables from the checkpoint.

        Returns
        -------
        Any
            The selected dynamic forcings variables.
        """
        return self.checkpoint.select_variables(
            **self.retrieved_dynamic_forcings_variables_include_exclude(),
        )

    def retrieved_dynamic_forcings_variables_and_mask(self) -> Any:
        """Select retrieved dynamic forcings variables and masks from the checkpoint.

        Returns
        -------
        Any
            The selected dynamic forcings variables and masks.
        """
        return self.checkpoint.select_variables_and_masks(
            **self.retrieved_dynamic_forcings_variables_include_exclude(),
        )

    ###############################################################################

    @classmethod
    def input_types(cls) -> dict[str, dict[str, list[str]]]:
        """Get all input types and their include/exclude lists.

        Returns
        -------
        dict
            Dictionary mapping input type names to include/exclude dicts.
        """
        return {
            "default-input": cls.default_runner_input_variables_include_exclude(),  # For backwards compatibility
            "constant-forcings": cls.retrieved_constant_forcings_variables_include_exclude(),
            "prognostics": cls.retrieved_prognostic_variables_include_exclude(),
            "dynamic-forcings": cls.retrieved_dynamic_forcings_variables_include_exclude(),
        }

    @classmethod
    def input_type_to_include_exclude(cls, input_type: str) -> dict[str, list[str]]:
        """Get the include/exclude dict for a given input type.

        Parameters
        ----------
        input_type : str
            The name of the input type.

        Returns
        -------
        dict
            The include/exclude dict for the input type.
        """
        return cls.input_types()[input_type]



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


from __future__ import annotations

import logging
import os
from copy import deepcopy
from functools import cached_property
from typing import Any
from typing import Literal

import numpy as np

from anemoi.inference.lazy import torch

from ..decorators import main_argument
from ..runners.default import DefaultRunner
from . import runner_registry

LOG = logging.getLogger(__name__)

# Possibly move the function(s) below to anemoi-models or anemoi-utils since it could be used in transfer learning.


def contains_any(key, specifications):
    contained = False
    for specification in specifications:
        if specification in key:
            contained = True
            break
    return contained


def update_state_dict(
    model, external_state_dict, keywords="", ignore_mismatched_layers=False, ignore_additional_layers=False
):
    """Update the model's stated_dict with entries from an external state_dict. Only entries whose keys contain the specified keywords are considered."""

    LOG.info("Updating model state dictionary.")

    if isinstance(keywords, str):
        keywords = [keywords]

    # select relevant part of external_state_dict
    reduced_state_dict = {k: v for k, v in external_state_dict.items() if contains_any(k, keywords)}
    model_state_dict = model.state_dict()

    # check layers and their shapes
    for key in list(reduced_state_dict):
        if key not in model_state_dict:
            if ignore_additional_layers:
                LOG.info("Skipping injection of %s, which is not in the model.", key)
                del reduced_state_dict[key]
            else:
                raise AssertionError(f"Layer {key} not in model. Consider setting 'ignore_additional_layers = True'.")
        elif reduced_state_dict[key].shape != model_state_dict[key].shape:
            if ignore_mismatched_layers:
                LOG.info("Skipping injection of %s due to shape mismatch.", key)
                LOG.info("Model shape: %s", model_state_dict[key].shape)
                LOG.info("Provided shape: %s", reduced_state_dict[key].shape)
                del reduced_state_dict[key]
            else:
                raise AssertionError(
                    "Mismatch in shape of %s. Consider setting 'ignore_mismatched_layers = True'.", key
                )

    # update
    model.load_state_dict(reduced_state_dict, strict=False)
    return model


def _get_supporting_arrays_from_graph(update_supporting_arrays: dict[str, str], graph: Any) -> dict:
    """Update the supporting arrays from the graph data."""
    updated_supporting_arrays = {}
    for key, value in update_supporting_arrays.items():
        if value in graph["data"]:
            updated_supporting_arrays[key] = graph["data"][value]
            LOG.info("Moving attribute '%s' from external graph to supporting arrays as '%s'.", value, key)
        else:
            error_msg = f"Key '{value}' not found in external graph 'data'. Cannot update supporting array: '{key}'."
            raise KeyError(error_msg)
    return updated_supporting_arrays


def _get_supporting_arrays_from_file(update_supporting_arrays: dict[str, str]) -> dict:
    """Update the supporting arrays from a file."""
    updated_supporting_arrays = {}

    for key, value in update_supporting_arrays.items():
        if os.path.isfile(value):
            try:
                updated_supporting_arrays[key] = torch.load(value)
            except Exception as e:
                LOG.warning("Failed to load '%s' as a torch tensor. Attempting to load as numpy array.\n%s", value, e)
                updated_supporting_arrays[key] = np.load(value, allow_pickle=True)

            LOG.info("Moving attribute '%s' from file to supporting arrays as '%s'.", value, key)
        else:
            error_msg = f"File '{value}' not found. Cannot update supporting array: '{key}'."
            raise FileNotFoundError(error_msg)
    return updated_supporting_arrays


def get_updated_supporting_arrays(
    update_supporting_arrays: dict[Literal["graph", "file"], dict[str, str]], graph: Any
) -> dict:
    """Get an update dict for the supporting arrays.

    Allows to update the supporting arrays in the checkpoint metadata from the graph data,
    and from files, which will be loaded using torch.load() or np.load().
    """
    updated_supporting_arrays = {}

    for location in update_supporting_arrays:
        if location == "graph":
            updated_supporting_arrays.update(
                _get_supporting_arrays_from_graph(update_supporting_arrays[location], graph)
            )
        elif location == "file":
            updated_supporting_arrays.update(_get_supporting_arrays_from_file(update_supporting_arrays[location]))
        else:
            raise ValueError(f"Unknown location '{location}' in update_supporting_arrays.")

    return updated_supporting_arrays


@runner_registry.register("external_graph")
@main_argument("graph")
class ExternalGraphRunner(DefaultRunner):
    """Runner where the graph saved in the checkpoint is replaced by an externally provided one.
    Currently only supported as an extension of the default runner.
    """

    def __init__(
        self,
        config: dict,
        graph: str,
        *,
        output_mask: dict | None = {},
        graph_dataset: Any | None = None,
        update_supporting_arrays: dict[Literal["graph", "file"], dict[str, str]] | None = None,
        updated_number_of_grid_points: str | int | None = None,
        check_state_dict: bool | None = True,
    ) -> None:
        """Initialise the ExternalGraphRunner.

        Parameters
        ----------
        config : Configuration
            Configuration for the runner.
        graph : str
            Path to the external graph.
        output_mask : dict | None
            Dictionary specifying the output mask.
        graph_dataset : Any | None
            Argument to open_dataset of anemoi-datasets that recreates the dataset used to build the data nodes of the graph.
        update_supporting_arrays: dict[Literal['graph', 'file'], dict[str, str]] | None
            Dictionary specifying how to update the supporting arrays in the checkpoint metadata.
            Allows both 'graph' and 'file' as keys.
            - 'graph': Will pull from the graph['data'] dictionary, key refers to the name in the supporting arrays,
            and value refers to the name in the graph['data'].
            - 'file': Will pull from the file system, key refers to the name in the supporting arrays,
            and value refers to the path to the file containing the numpy array.
                Can be torch.load() or np.load() compatible.
        updated_number_of_grid_points: str | int | None
            If the number of grid points in the graph is different from that in the checkpoint,
            this can be used to update the number of grid points in the checkpoint metadata.
            If is a str, will be used as a key to the graph['data'] dictionary.
            If is an int, will be used as the number of grid points.
        check_state_dict: bool | None
            Boolean specifying if reconstruction of statedict happens as expeceted.
        """
        super().__init__(config)
        self.check_state_dict = check_state_dict
        self.graph_path = graph

        # If graph was build on other dataset, we need to adapt the dataloader
        if graph_dataset is not None:
            from anemoi.datasets import open_dataset

            graph_ds = open_dataset(graph_dataset)
            LOG.info(
                "The external graph was built using a different anemoi-dataset than that in the checkpoint. "
                "Patching metadata to ensure correct data loading."
            )
            self.checkpoint._metadata.patch(
                {
                    "config": {"dataloader": {"dataset": graph_dataset}},
                    "dataset": {"shape": graph_ds.shape},
                }
            )

            # had to use private attributes because cached properties cause problems
            self.checkpoint._metadata._supporting_arrays.update(graph_ds.supporting_arrays())
            if "grid_indices" in self.checkpoint._metadata._supporting_arrays:
                num_grid_points = len(self.checkpoint._metadata._supporting_arrays["grid_indices"])
            else:
                num_grid_points = graph_ds.shape[-1]
            self.checkpoint._metadata.number_of_grid_points = num_grid_points

        # Check if the external graph has the 'indices_connected_nodes' attribute
        # If so adapt dataloader and add supporting array
        data = self.checkpoint._metadata._config.graph.data
        assert data in self.graph.node_types, f"Node type {data} not found in external graph."
        if "indices_connected_nodes" in self.graph[data]:
            LOG.info(
                "The external graph has the 'indices_connected_nodes' attribute."
                "Patching metadata with MaskedGrid 'grid_indices' to ensure correct data loading."
            )
            self.checkpoint._metadata.patch(
                {
                    "config": {
                        "dataloader": {
                            "grid_indices": {
                                "_target_": "anemoi.training.data.grid_indices.MaskedGrid",
                                "nodes_name": data,
                                "node_attribute_name": "indices_connected_nodes",
                            }
                        }
                    }
                }
            )
            LOG.info("Moving 'indices_connected_nodes' from external graph to supporting arrays as 'grid_indices'.")
            indices_connected_nodes = self.graph[data]["indices_connected_nodes"].numpy()
            self.checkpoint._supporting_arrays["grid_indices"] = indices_connected_nodes.squeeze()

        if output_mask:
            nodes = output_mask["nodes_name"]
            attribute = output_mask["attribute_name"]
            self.checkpoint._supporting_arrays["output_mask"] = self.graph[nodes][attribute].numpy().squeeze()
            LOG.info(
                "Moving attribute '%s' of nodes '%s' from external graph to supporting arrays as 'output_mask'.",
                attribute,
                nodes,
            )

        if update_supporting_arrays is not None:
            self.checkpoint._supporting_arrays.update(
                get_updated_supporting_arrays(update_supporting_arrays, self.graph)
            )

        if updated_number_of_grid_points is not None:
            if isinstance(updated_number_of_grid_points, str):
                updated_number_of_grid_points = len(self.graph["data"][updated_number_of_grid_points])
            self.checkpoint._metadata.number_of_grid_points = updated_number_of_grid_points
            LOG.info(
                "Updated number of grid points in the checkpoint metadata to %s.",
                updated_number_of_grid_points,
            )

    @cached_property
    def graph(self):

        graph_path = self.graph_path
        assert os.path.isfile(
            graph_path
        ), f"No graph found at {graph_path}. An external graph needs to be specified in the config file for this runner."
        LOG.info("Loading external graph from path %s.", graph_path)
        return torch.load(graph_path, map_location="cpu", weights_only=False)

    @cached_property
    def model(self):
        # load the model from the checkpoint
        device = self.device
        self.device = "cpu"
        model_instance = super().model
        state_dict_ckpt = deepcopy(model_instance.state_dict())

        # rebuild the model with the new graph
        model_instance.graph_data = self.graph
        model_instance.config = self.checkpoint._metadata._config
        model_instance._build_model()

        # reinstate the weights, biases and normalizer from the checkpoint
        # reinstating the normalizer is necessary for checkpoints that were created
        # using transfer learning, where the statistics as stored in the checkpoint
        # do not match the statistics used to build the normalizer in the checkpoint.
        model_instance = update_state_dict(
            model_instance, state_dict_ckpt, keywords=["bias", "weight", "processors.normalizer"]
        )

        LOG.info("Successfully built model with external graph and reassigned model weights!")
        self.device = device
        return model_instance.to(self.device)



# (C) Copyright 2025 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import json
import logging
from typing import Any

import earthkit.data as ekd

from anemoi.inference.types import State

from . import create_template_provider

LOG = logging.getLogger(__name__)


class TemplateManager:
    """A class to manage GRIB templates."""

    def __init__(self, owner: Any, templates: list[str] | str | None = None) -> None:
        """Initialize the TemplateManager.

        Parameters
        ----------
        owner : Any
            The owner of the TemplateManager.
        templates : Optional[Union[List[str], str]], optional
            A list of template names or a single template name, by default None.
        """
        self.owner = owner
        self.checkpoint = owner.context.checkpoint
        self.typed_variables = self.checkpoint.typed_variables

        self._template_cache = {}

        if templates is None:
            templates = []

        if not isinstance(templates, (list, tuple)):
            templates = [templates]

        if len(templates) == 0:
            templates = ["builtin"]

        self.templates_providers = [create_template_provider(self, template) for template in templates]

    def template(self, name: str, state: State, typed_variables: dict[str, Any]) -> ekd.Field | None:
        """Get the template for a given name and state.

        Parameters
        ----------
        name : str
            The name of the template.
        state : State
            The state object containing template information.
        typed_variables : dict[str, Any]
            The dictionary of typed variables.

        Returns
        -------
        Optional[ekd.Field]
            The template field if found, otherwise None.
        """
        assert name is not None, name

        # Use input fields as templates
        self._template_cache.update(state.get("_grib_templates_for_output", {}))

        if name not in self._template_cache:
            self.load_template(name, state, typed_variables)

        return self._template_cache.get(name)

    def load_template(self, name: str, state: State, typed_variables: dict[str, Any]) -> ekd.Field | None:
        """Load the template for a given name and state.

        Parameters
        ----------
        name : str
            The name of the template.
        state : State
            The state object containing template information.
        typed_variables : dict[str, Any]
            The dictionary of typed variables.

        Returns
        -------
        Optional[ekd.Field]
            The template field if found, otherwise None.
        """

        checkpoint = self.owner.context.checkpoint

        typed = typed_variables[name]

        grid = state.get("_geography", {}).get("grid", checkpoint.grid)
        area = state.get("_geography", {}).get("area", checkpoint.area)

        lookup = dict(
            name=name,
            grid=self._grid(grid),
            area=area,
            time_processing=typed.time_processing,
            number_of_grid_points=checkpoint.number_of_grid_points,
        )

        for key, value in typed.grib_keys.items():
            if key in ("step", "date", "time", "hdate"):
                continue
            lookup[key] = value

        lookup.update(self.owner.template_lookup(name))

        if LOG.isEnabledFor(logging.DEBUG):
            LOG.debug(f"Loading template for `{name}` with lookup:")
            LOG.debug("%s", json.dumps(lookup, indent=2, default=str))

        tried = []
        for provider in self.templates_providers:
            template = provider.template(name, lookup)
            if template is not None:
                self._template_cache[name] = template
                return

            tried.append(provider)

        LOG.warning(f"Could not find template for `{name}` in {tried}")
        LOG.warning(f"Loading template for `{name}` with lookup:")
        LOG.warning("%s", json.dumps(lookup, indent=2, default=str))
        return None

    def _grid(self, grid: str | list[float] | tuple[int, int]) -> str:
        """Convert the grid information to a standardised format.

        Parameters
        ----------
        grid : Union[str, list of float, tuple of int]
            The grid information.

        Returns
        -------
        str
            The standardised grid format.
        """

        if isinstance(grid, str):
            return grid.upper()

        if isinstance(grid, (tuple, list)) and len(grid) == 2:
            if grid[0] == grid[1]:
                return grid[0]
            return f"{grid[0]}x{grid[1]}"

        return grid



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging

import numpy as np

from anemoi.inference.config import Configuration
from anemoi.inference.context import Context
from anemoi.inference.types import ProcessorConfig

from . import output_registry
from .masked import MaskedOutput

LOG = logging.getLogger(__name__)


@output_registry.register("extract_lam")
class ExtractLamOutput(MaskedOutput):
    """Extract LAM output class."""

    def __init__(
        self,
        context: Context,
        *,
        output: Configuration,
        lam: str = "lam_0",
        variables: list[str] | None = None,
        post_processors: list[ProcessorConfig] | None = None,
        output_frequency: int | None = None,
        write_initial_state: bool | None = None,
    ) -> None:
        """Parameters
        ----------
        context : dict
            The context dictionary.
        output : dict
            The output configuration dictionary.
        lam : str, optional
            The LAM identifier, by default "lam_0".
        variables : list, optional
            The list of variables to extract, by default None.
        post_processors : Optional[List[ProcessorConfig]], default None
            Post-processors to apply to the input
        output_frequency : int, optional
            The frequency of output, by default None.
        write_initial_state : bool, optional
            Whether to write the initial state, by default None.
        """

        if "cutout_mask" in context.checkpoint.supporting_arrays:
            # Backwards compatibility
            mask = context.checkpoint.load_supporting_array("cutout_mask")
            points = slice(None, -np.sum(mask))
        else:
            if "lam_0" not in lam:
                raise NotImplementedError("Only lam_0 is supported")

            if "lam_1/cutout_mask" in context.checkpoint.supporting_arrays:
                raise NotImplementedError("Only lam_0 is supported")

            mask = context.checkpoint.load_supporting_array(f"{lam}/cutout_mask")

            assert len(mask) == np.sum(mask)
            points = slice(None, np.sum(mask))

        super().__init__(
            context,
            mask=points,
            output=output,
            variables=variables,
            post_processors=post_processors,
            output_frequency=output_frequency,
            write_initial_state=write_initial_state,
        )



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from pathlib import Path
from typing import Any
from typing import TypeVar

from anemoi.inference.context import Context

LOG = logging.getLogger("anemoi.inference")

MARKER = object()
F = TypeVar("F", bound=type)


class main_argument:
    """Decorator to set the main argument of a class. Only for classes with a 'context' argument.

    For example:
    ```
    @main_argument("path")
    class GribOutput
        def __init__(context, encoding=None, path=None, archive_requests=None):
            ...
    output = GribOutput(context, "out.grib")
    ```
    So in the config we can have:
    ```
    output:
        grib: out.grib
    ```
    meaning the same as
    ```
    output:
        grib:
            path: out.grib
    ```
    """

    def __init__(self, name: str):
        """Initialize the main_argument decorator.

        Parameters
        ----------
        name : str
            The name of the main argument.
        """
        self.name = name

    def __call__(self, cls: F) -> F:
        """Decorate the class to set the main argument."""

        if not isinstance(cls, type):
            raise TypeError("'main_argument' can only be used to decorate classes")

        class WrappedClass(cls):
            def __init__(wrapped_cls, context: Context, main: object = MARKER, *args: Any, **kwargs: Any) -> Any:
                if main is not MARKER:
                    kwargs[self.name] = main
                super().__init__(context, *args, **kwargs)

        return type(cls.__name__, (WrappedClass,), {})


class ensure_path:
    """Decorator to ensure a path argument is a Path object and optionally exists.

    If `is_dir` is True, the path is treated as a directory, if not for files, the parent directory is treated as a directory.
    If `must_exist` is True, the directory must exist.
    If `create` is True, the directory will be created if it doesn't exist.

    For example:
    ```
    @ensure_path("dir", create=True)
    class GribOutput
        def __init__(context, dir=None, archive_requests=None):
            ...
    """

    def __init__(self, arg: str, is_dir: bool = False, create: bool = True, must_exist: bool = False):
        self.arg = arg
        self.is_dir = is_dir
        self.create = create
        self.must_exist = must_exist

    def __call__(self, cls: F) -> F:
        """Decorate the object to ensure the path argument is a Path object."""

        class WrappedClass(cls):
            def __init__(wrapped_cls, context: Context, *args: Any, **kwargs: Any) -> None:
                if self.arg not in kwargs:
                    LOG.debug(f"Argument '{self.arg}' not found in kwargs, cannot ensure path.")
                    super().__init__(context, *args, **kwargs)
                    return

                path = kwargs[self.arg] = Path(kwargs[self.arg])
                if not self.is_dir:
                    path = path.parent

                if self.must_exist:
                    if not path.exists():
                        raise FileNotFoundError(f"Path '{path}' does not exist.")
                if self.create:
                    path.mkdir(parents=True, exist_ok=True)

                super().__init__(context, *args, **kwargs)

        return type(cls.__name__, (WrappedClass,), {})


class ensure_dir(ensure_path):
    """Decorator to ensure a directory path argument is a Path object and optionally exists.

    If `must_exist` is True, the directory must exist.
    If `create` is True, the directory will be created if it doesn't exist.

    For example:
    ```
    @ensure_dir("dir", create=True)
    class PlotOutput
        def __init__(context, dir=None, ...):
            ...
    """

    def __init__(self, arg: str, create: bool = True, must_exist: bool = False):
        super().__init__(arg, is_dir=True, create=create, must_exist=must_exist)
